from typing import Optional, List, Any, Mapping

from gptcache.adapter.adapter import adapt, aadapt
from gptcache.core import cache
from gptcache.manager.scalar_data.base import Answer, DataType
from gptcache.session import Session
from gptcache.utils import import_langchain

import_langchain()

# pylint: disable=C0413
from langchain.llms.base import LLM
from langchain.chat_models.base import BaseChatModel
from langchain.schema import (
    BaseMessage,
    LLMResult,
    AIMessage,
    ChatGeneration,
    ChatResult,
)
from langchain.callbacks.manager import (
    Callbacks,
    CallbackManagerForLLMRun, AsyncCallbackManagerForLLMRun,
)


# pylint: disable=protected-access
class LangChainLLMs(LLM):
    """LangChain LLM Wrapper.

    :param llm: LLM from langchain.llms.
    :type llm: Any

    Example:
        .. code-block:: python

            from gptcache import cache
            from gptcache.processor.pre import get_prompt
            # init gptcache
            cache.init(pre_embedding_func=get_prompt)
            cache.set_openai_key()

            from langchain.llms import OpenAI
            from gptcache.adapter.langchain_models import LangChainLLMs
            # run llm with gptcache
            llm = LangChainLLMs(llm=OpenAI(temperature=0))
            llm("Hello world")
    """

    llm: Any
    session: Session = None
    tmp_args: Any = None

    @property
    def _llm_type(self) -> str:
        return self.llm._llm_type

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        return self.llm._identifying_params

    def __str__(self) -> str:
        return str(self.llm)

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        _: Optional[CallbackManagerForLLMRun] = None,
    ) -> str:
        session = (
            self.session
            if "session" not in self.tmp_args
            else self.tmp_args.pop("session")
        )
        cache_obj = self.tmp_args.pop("cache_obj", cache)
        return adapt(
            self.llm,
            _cache_data_convert,
            _update_cache_callback,
            prompt=prompt,
            stop=stop,
            cache_obj=cache_obj,
            session=session,
            **self.tmp_args,
        )

    async def _acall(self, prompt: str, stop: Optional[List[str]] = None,
                     run_manager: Optional[AsyncCallbackManagerForLLMRun] = None) -> str:
        return await super()._acall(prompt, stop=stop, run_manager=run_manager)

    def generate(
        self,
        prompts: List[str],
        stop: Optional[List[str]] = None,
        callbacks: Callbacks = None,
        **kwargs,
    ) -> LLMResult:
        self.tmp_args = kwargs
        return super().generate(prompts, stop=stop, callbacks=callbacks)

    async def agenerate(
        self,
        prompts: List[str],
        stop: Optional[List[str]] = None,
        callbacks: Callbacks = None,
        **kwargs,
    ) -> LLMResult:
        self.tmp_args = kwargs
        return await super().agenerate(prompts, stop=stop, callbacks=callbacks)

    def __call__(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        callbacks: Callbacks = None,
        **kwargs,
    ) -> str:
        """Check Cache and run the LLM on the given prompt and input."""
        return (
            self.generate([prompt], stop=stop, callbacks=callbacks, **kwargs)
            .generations[0][0]
            .text
        )


# pylint: disable=protected-access
class LangChainChat(BaseChatModel):
    """LangChain LLM Wrapper.

    :param chat: LLM from langchain.chat_models.
    :type chat: Any

    Example:
        .. code-block:: python

            from gptcache import cache
            from gptcache.processor.pre import get_messages_last_content
            # init gptcache
            cache.init(pre_embedding_func=get_messages_last_content)
            cache.set_openai_key()
            from langchain.chat_models import ChatOpenAI
            from gptcache.adapter.langchain_models import LangChainChat
            # run chat model with gptcache
            chat = LangChainChat(chat=ChatOpenAI(temperature=0))
            chat([HumanMessage(content="Translate this sentence from English to French. I love programming.")])
    """

    @property
    def _llm_type(self) -> str:
        return "gptcache_llm_chat"

    chat: Any
    session: Optional[Session] = None
    tmp_args: Optional[Any] = None

    def _generate(
        self,
        messages: Any,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
    ) -> ChatResult:
        session = (
            self.session
            if "session" not in self.tmp_args
            else self.tmp_args.pop("session")
        )
        cache_obj = self.tmp_args.pop("cache_obj", cache)
        return adapt(
            self.chat._generate,
            _cache_msg_data_convert,
            _update_cache_msg_callback,
            messages=messages,
            stop=stop,
            cache_obj=cache_obj,
            session=session,
            run_manager=run_manager,
            **self.tmp_args,
        )

    async def _agenerate(
        self,
        messages: List[List[BaseMessage]],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
    ) -> ChatResult:
        session = (
            self.session
            if "session" not in self.tmp_args
            else self.tmp_args.pop("session")
        )
        cache_obj = self.tmp_args.pop("cache_obj", cache)
        return await aadapt(
            self.chat._agenerate,
            _cache_msg_data_convert,
            _update_cache_msg_callback,
            messages=messages,
            stop=stop,
            cache_obj=cache_obj,
            session=session,
            run_manager=run_manager,
            **self.tmp_args,
        )

    def generate(
        self,
        messages: List[List[BaseMessage]],
        stop: Optional[List[str]] = None,
        callbacks: Callbacks = None,
        **kwargs,
    ) -> LLMResult:
        self.tmp_args = kwargs
        return super().generate(messages, stop=stop, callbacks=callbacks)

    async def agenerate(
        self,
        messages: List[List[BaseMessage]],
        stop: Optional[List[str]] = None,
        callbacks: Callbacks = None,
        **kwargs,
    ) -> LLMResult:
        self.tmp_args = kwargs
        return await super().agenerate(messages, stop=stop, callbacks=callbacks)

    @property
    def _identifying_params(self):
        return self.chat._identifying_params

    def _combine_llm_outputs(self, llm_outputs: List[Optional[dict]]) -> dict:
        return self.chat._combine_llm_outputs(llm_outputs)

    def get_num_tokens(self, text: str) -> int:
        return self.chat.get_num_tokens(text)

    def get_num_tokens_from_messages(self, messages: List[BaseMessage]) -> int:
        return self.chat.get_num_tokens_from_messages(messages)

    def __call__(self, messages: Any, stop: Optional[List[str]] = None, **kwargs):
        generation = self.generate([messages], stop=stop, **kwargs).generations[0][0]
        if isinstance(generation, ChatGeneration):
            return generation.message
        else:
            raise ValueError("Unexpected generation type")


def _cache_data_convert(cache_data):
    return cache_data


def _update_cache_callback(
    llm_data, update_cache_func, *args, **kwargs
):  # pylint: disable=unused-argument
    update_cache_func(Answer(llm_data, DataType.STR))
    return llm_data


def _cache_msg_data_convert(cache_data):
    llm_res = ChatResult(
        generations=[
            ChatGeneration(
                text="",
                generation_info=None,
                message=AIMessage(content=cache_data, additional_kwargs={}),
            )
        ],
        llm_output=None,
    )
    return llm_res


def _update_cache_msg_callback(
    llm_data, update_cache_func, *args, **kwargs
):  # pylint: disable=unused-argument
    update_cache_func(llm_data.generations[0].text)
    return llm_data



__all__ = [
    "OpenAI",
    "Huggingface",
    "SBERT",
    "Cohere",
    "Onnx",
    "FastText",
    "Data2VecAudio",
    "Timm",
    "ViT",
    "LangChain",
    "Rwkv",
    "PaddleNLP",
    "UForm",
]


from gptcache.utils.lazy_import import LazyImport

openai = LazyImport("openai", globals(), "gptcache.embedding.openai")
huggingface = LazyImport("huggingface", globals(), "gptcache.embedding.huggingface")
sbert = LazyImport("sbert", globals(), "gptcache.embedding.sbert")
onnx = LazyImport("onnx", globals(), "gptcache.embedding.onnx")
cohere = LazyImport("cohere", globals(), "gptcache.embedding.cohere")
fasttext = LazyImport("fasttext", globals(), "gptcache.embedding.fasttext")
data2vec = LazyImport("data2vec", globals(), "gptcache.embedding.data2vec")
timm = LazyImport("timm", globals(), "gptcache.embedding.timm")
vit = LazyImport("vit", globals(), "gptcache.embedding.vit")
langchain = LazyImport("langchain", globals(), "gptcache.embedding.langchain")
rwkv = LazyImport("rwkv", globals(), "gptcache.embedding.rwkv")
paddlenlp = LazyImport("paddlenlp", globals(), "gptcache.embedding.paddlenlp")
uform = LazyImport("uform", globals(), "gptcache.embedding.uform")


def Cohere(model="large", api_key=None):
    return cohere.Cohere(model, api_key)


def OpenAI(model="text-embedding-ada-002", api_key=None):
    return openai.OpenAI(model, api_key)


def Huggingface(model="distilbert-base-uncased"):
    return huggingface.Huggingface(model)


def SBERT(model="all-MiniLM-L6-v2"):
    return sbert.SBERT(model)


def Onnx(model="GPTCache/paraphrase-albert-onnx"):
    return onnx.Onnx(model)


def FastText(model="en", dim=None):
    return fasttext.FastText(model, dim)


def Data2VecAudio(model="facebook/data2vec-audio-base-960h"):
    return data2vec.Data2VecAudio(model)


def Timm(model="resnet50", device="default"):
    return timm.Timm(model, device)


def ViT(model="google/vit-base-patch16-384"):
    return vit.ViT(model)


def LangChain(embeddings, dimension=0):
    return langchain.LangChain(embeddings, dimension)


def Rwkv(model="sgugger/rwkv-430M-pile"):
    return rwkv.Rwkv(model)


def PaddleNLP(model="ernie-3.0-medium-zh"):
    return paddlenlp.PaddleNLP(model)


def UForm(model="unum-cloud/uform-vl-multilingual", embedding_type="text"):
    return uform.UForm(model, embedding_type)



from typing import List, Optional, Tuple

import numpy as np

from gptcache.manager.vector_data.base import VectorBase, VectorData
from gptcache.utils import import_docarray

import_docarray()

from docarray.typing import NdArray  # pylint: disable=C0413
from docarray import BaseDoc, DocList  # pylint: disable=C0413
from docarray.index import InMemoryExactNNIndex  # pylint: disable=C0413


class DocarrayVectorData(BaseDoc):
    """Class representing a vector data element with an ID and associated data."""

    id: int
    data: NdArray


class DocArrayIndex(VectorBase):
    """
    Class representing in-memory exact nearest neighbor index for vector search.

    :param index_file_path: the path to docarray index, defaults to 'docarray_index.bin'.
    :type index_file_path: str
    :param top_k: the number of the vectors results to return, defaults to 1.
    :type top_k: int
    """

    def __init__(self, index_file_path: str, top_k: int):
        self._index = InMemoryExactNNIndex[DocarrayVectorData](
            index_file_path=index_file_path
        )
        self._index_file_path = index_file_path
        self._top_k = top_k

    def mul_add(self, datas: List[VectorData]) -> None:
        """
        Add multiple vector data elements to the index.

        :param datas: A list of vector data elements to be added.
        """
        docs = DocList[DocarrayVectorData](
            DocarrayVectorData(id=data.id, data=data.data) for data in datas
        )
        self._index.index(docs)

    def search(
        self, data: np.ndarray, top_k: int = -1
    ) -> Optional[List[Tuple[float, int]]]:
        """
        Search for the nearest vector data elements in the index.

        :param data: The query vector data.
        :param top_k: The number of top matches to return.
        :return: A list of tuples, each containing the match score and
            the ID of the matched vector data element.
        """
        if len(self._index) == 0:
            return None
        if top_k == -1:
            top_k = self._top_k
        docs, scores = self._index.find(data, search_field="data", limit=top_k)
        return list(zip(scores, docs.id))

    def rebuild(self, ids: Optional[List[int]] = None) -> bool:
        """
        In the case of DocArrayIndex, the rebuild operation is not needed.
        """
        return True

    def delete(self, ids: Optional[List[str]]) -> None:
        """
        Delete the specified vector data elements from the index.

        :param ids: A list of IDs of the vector data elements to be deleted.
        """
        if ids is not None:
            del self._index[ids]

    def flush(self) -> None:
        self._index.persist(self._index_file_path)

    def close(self) -> None:
        self.flush()



from typing import Any, Dict

from gptcache.processor import ContextProcess
from gptcache.utils import import_selective_context

import_selective_context()

from selective_context import SelectiveContext  # pylint: disable=C0413


class SelectiveContextProcess(ContextProcess):
    """A context processor for selecting context

    Need to download the corresponding model before use, the default English model is: en_core_web_sm

    `pip install spacy && python -m spacy download en_core_web_sm`

    :param model_type: the selective context model name, default value is 'gpt2'
    :type model_type: str
    :param lang: the content lang type, default value is 'en'.
    :type lang: str
    :param reduce_ratio: selective context ratio. The range for the value is between 0 and 1, with a default value of 0.35.
    :type reduce_ratio: float
    :param reduce_level: selective context level. The valid values include 'sent', 'phrase', and 'token', with the default value being 'phrase'.
    :type reduce_level: str

    more details: https://github.com/liyucheng09/Selective_Context

    Example:
        .. code-block:: python

            from gptcache.processor.context.selective_context import SelectiveContextProcess

            context_process = SelectiveContextProcess()
            cache.init(pre_embedding_func=context_process.pre_process)
    """

    content: str = ""

    def __init__(
            self,
            model_type: str = "gpt2",
            lang: str = "en",
            reduce_ratio: float = 0.35,
            reduce_level: str = "phrase",
    ):
        self.sc = SelectiveContext(model_type=model_type, lang=lang)
        self.reduce_ratio = reduce_ratio
        self.reduce_level = reduce_level

    def format_all_content(self, data: Dict[str, Any], **params: Dict[str, Any]):
        for query in data["messages"]:
            self.content += f"{query['role']}: {query['content']} \n"

    def process_all_content(self) -> (Any, Any):
        selective_content, _ = self.sc(
            self.content, reduce_ratio=self.reduce_ratio, reduce_level=self.reduce_level
        )
        return self.content, selective_content



import subprocess

from gptcache.utils.error import PipInstallError
from gptcache.utils.log import gptcache_log


def prompt_install(package: str, warn: bool = False):  # pragma: no cover
    """
    Function used to prompt user to install a package.
    """
    cmd = f"pip install -q {package}"
    try:
        if warn and input(f"Install {package}? Y/n: ") != "Y":
            raise ModuleNotFoundError(f"No module named {package}")
        print(f"start to install package: {package}")
        subprocess.check_call(cmd, shell=True)
        print(f"successfully installed package: {package}")
        gptcache_log.info("%s installed successfully!", package)
    except subprocess.CalledProcessError as e:
        raise PipInstallError(package) from e



import uuid
from typing import Callable, Optional

from gptcache import cache
from gptcache.manager.data_manager import DataManager
from gptcache.processor.check_hit import check_hit_session
from gptcache.utils.log import gptcache_log


class Session:
    """
    Session for gptcache. Session can isolate the context of each connection, and can also filter the results after recall,
    and if not satisfied will re-request rather than return the cache results directly.

    :param name: the name of the session, defaults to `uuid.uuid4().hex`.
    :type name:  str
    :param data_manager: the DataManager of the session, defaults to cache.data_manager with the initialized cache.
    :type data_manager: DataManager
    :param check_hit_func: a Callable to check the hit, defaults to `processor.check_hit.check_hit_session`，which will not return cached data
                           if you ask the same or similar question in the same session.
    :type check_hit_func:  Callable


    Example:
        .. code-block:: python

            from gptcache import cache
            from gptcache.session import Session
            # init gptcache
            cache.init()
            cache.set_openai_key()
            session = Session()

            from gptcache.adapter import openai
            # run ChatCompletion model with gptcache on session
            response = openai.ChatCompletion.create(
                          model='gpt-3.5-turbo',
                          messages=[
                            {
                                'role': 'user',
                                'content': "what's github"
                            }],
                          session=session
                        )
            response_content = response['choices'][0]['message']['content']
    """

    def __init__(
        self,
        name: Optional[str] = None,
        data_manager: Optional[DataManager] = None,
        check_hit_func: Optional[Callable] = None,
    ):
        self._name = uuid.uuid4().hex if not name else name
        self._data_manager = cache.data_manager if not data_manager else data_manager
        self.check_hit_func = (
            check_hit_session if not check_hit_func else check_hit_func
        )

    @property
    def name(self):
        return self._name

    def __enter__(self):
        gptcache_log.warning(
            "The `with` method will delete the session data directly on exit."
        )
        return self

    def __exit__(self, *_):
        self.drop()

    def drop(self):
        """Drop the session and delete all data in the session"""
        self._data_manager.delete_session(self.name)
        gptcache_log.info("Deleting data in the session: %s.", self.name)



from gptcache.utils.error import NotFoundError, ParamError

TOP_K = 1

FAISS_INDEX_PATH = "faiss.index"
DIMENSION = 0

MILVUS_HOST = "localhost"
MILVUS_PORT = 19530
MILVUS_USER = ""
MILVUS_PSW = ""
MILVUS_SECURE = False
MILVUS_INDEX_PARAMS = {
    "metric_type": "L2",
    "index_type": "HNSW",
    "params": {"M": 8, "efConstruction": 64},
}

PGVECTOR_URL = "postgresql://postgres:postgres@localhost:5432/postgres"
PGVECTOR_INDEX_PARAMS = {"index_type": "L2", "params": {"lists": 100, "probes": 10}}

QDRANT_GRPC_PORT = 6334
QDRANT_HTTP_PORT = 6333
QDRANT_INDEX_PARAMS = {"ef_construct": 100, "m": 16}
QDRANT_DEFAULT_LOCATION = "./qdrant_data"
QDRANT_FLUSH_INTERVAL_SEC = 5

COLLECTION_NAME = "gptcache"

WEAVIATE_TIMEOUT_CONFIG = (10, 60)
WEAVIATE_STARTUP_PERIOD = 5


# pylint: disable=import-outside-toplevel
class VectorBase:
    """
    VectorBase to manager the vector base.

    Generate specific VectorBase with the configuration. For example, setting for
       `Milvus` (with , `host`, `port`, `password`, `secure`, `collection_name`, `index_params`, `search_params`, `local_mode`, `local_data` params),
       `Faiss` (with , `index_path`, `dimension`, `top_k` params),
       `Chromadb` (with `top_k`, `client_settings`, `persist_directory`, `collection_name` params),
       `Hnswlib` (with `index_file_path`, `dimension`, `top_k`, `max_elements` params).
       `pgvector` (with `url`, `collection_name`, `index_params`, `top_k`, `dimension` params).

    :param name: the name of the vectorbase, it is support 'milvus', 'faiss', 'chromadb', 'hnswlib' now.
    :type name: str

    :param top_k: the number of the vectors results to return, defaults to 1.
    :type top_k: int

    :param dimension: the dimension of the vector, defaults to 0.
    :type dimension: int

    :param index_path: the path to Faiss index, defaults to 'faiss.index'.
    :type index_path: str

    :param host: the host for Milvus vector database, defaults to 'localhost'.
    :type host: str
    :param port: the port for Milvus vector database, defaults to '19530'.
    :type port: str
    :param user: the user for Zilliz Cloud, defaults to "".
    :type user: str
    :param password: the password for Zilliz Cloud, defaults to "".
    :type password: str
    :param secure: whether it is https with Zilliz Cloud, defaults to False.
    :type secures: bool
    :param index_params: the index parameters for Milvus, defaults to the HNSW index: {'metric_type': 'L2', 'index_type': 'HNSW', 'params': {'M':
                         8, 'efConstruction': 64}}.
    :type index_params: dict
    :param search_params: the index parameters for Milvus, defaults to None.
    :type search_params: dict
    :param collection_name: the name of the collection for Milvus vector database, defaults to 'gptcache'.
    :type collection_name: str
    :param local_mode: if true, will start a local milvus server.
    :type local_mode: bool
    :param local_data: required when local_mode is True.
    :type local_data: str

    :param url: the connection url for PostgreSQL database, defaults to 'postgresql://postgres@localhost:5432/postgres'
    :type url: str
    :param index_params: the index parameters for pgvector.
    :type index_params: dict
    :param collection_name: the prefix of the table for PostgreSQL pgvector, defaults to 'gptcache'.
    :type collection_name: str

    :param client_settings: the setting for Chromadb.
    :type client_settings: Settings
    :param persist_directory: the directory to persist, defaults to '.chromadb/' in the current directory.
    :type persist_directory: str

    :param index_path: the path to hnswlib index, defaults to 'hnswlib_index.bin'.
    :type index_path: str
    :param max_elements: max_elements of hnswlib, defaults 100000.
    :type max_elements: int
    """

    def __init__(self):
        raise EnvironmentError(
            "VectorBase is designed to be instantiated, please using the `VectorBase.get(name)`."
        )

    @staticmethod
    def check_dimension(dimension):
        if dimension <= 0:
            raise ParamError(
                f"the dimension should be greater than zero, current value: {dimension}."
            )

    @staticmethod
    def get(name, **kwargs):
        top_k = kwargs.get("top_k", TOP_K)
        if name == "milvus":
            from gptcache.manager.vector_data.milvus import Milvus

            dimension = kwargs.get("dimension", DIMENSION)
            VectorBase.check_dimension(dimension)
            host = kwargs.get("host", MILVUS_HOST)
            port = kwargs.get("port", MILVUS_PORT)
            user = kwargs.get("user", MILVUS_USER)
            password = kwargs.get("password", MILVUS_PSW)
            secure = kwargs.get("secure", MILVUS_SECURE)
            collection_name = kwargs.get("collection_name", COLLECTION_NAME)
            index_params = kwargs.get("index_params", MILVUS_INDEX_PARAMS)
            search_params = kwargs.get("search_params", None)
            local_mode = kwargs.get("local_mode", False)
            local_data = kwargs.get("local_data", "./milvus_data")
            vector_base = Milvus(
                host=host,
                port=port,
                user=user,
                password=password,
                secure=secure,
                collection_name=collection_name,
                dimension=dimension,
                top_k=top_k,
                index_params=index_params,
                search_params=search_params,
                local_mode=local_mode,
                local_data=local_data,
            )
        elif name == "faiss":
            from gptcache.manager.vector_data.faiss import Faiss

            dimension = kwargs.get("dimension", DIMENSION)
            index_path = kwargs.pop("index_path", FAISS_INDEX_PATH)
            VectorBase.check_dimension(dimension)
            vector_base = Faiss(
                index_file_path=index_path, dimension=dimension, top_k=top_k
            )
        elif name == "chromadb":
            from gptcache.manager.vector_data.chroma import Chromadb

            client_settings = kwargs.get("client_settings", None)
            persist_directory = kwargs.get("persist_directory", None)
            collection_name = kwargs.get("collection_name", COLLECTION_NAME)
            vector_base = Chromadb(
                client_settings=client_settings,
                persist_directory=persist_directory,
                collection_name=collection_name,
                top_k=top_k,
            )
        elif name == "hnswlib":
            from gptcache.manager.vector_data.hnswlib_store import Hnswlib

            dimension = kwargs.get("dimension", DIMENSION)
            index_path = kwargs.pop("index_path", "./hnswlib_index.bin")
            max_elements = kwargs.pop("max_elements", 100000)
            VectorBase.check_dimension(dimension)
            vector_base = Hnswlib(
                index_file_path=index_path,
                dimension=dimension,
                top_k=top_k,
                max_elements=max_elements,
            )
        elif name == "pgvector":
            from gptcache.manager.vector_data.pgvector import PGVector

            dimension = kwargs.get("dimension", DIMENSION)
            url = kwargs.get("url", PGVECTOR_URL)
            collection_name = kwargs.get("collection_name", COLLECTION_NAME)
            index_params = kwargs.get("index_params", PGVECTOR_INDEX_PARAMS)
            vector_base = PGVector(
                dimension=dimension,
                top_k=top_k,
                url=url,
                collection_name=collection_name,
                index_params=index_params,
            )
        elif name == "docarray":
            from gptcache.manager.vector_data.docarray_index import DocArrayIndex

            index_path = kwargs.pop("index_path", "./docarray_index.bin")
            vector_base = DocArrayIndex(index_file_path=index_path, top_k=top_k)
        elif name == "usearch":
            from gptcache.manager.vector_data.usearch import USearch

            dimension = kwargs.get("dimension", DIMENSION)
            index_path = kwargs.pop("index_path", "./index.usearch")
            metric = kwargs.get("metric", "cos")
            dtype = kwargs.get("dtype", "f32")
            vector_base = USearch(
                index_file_path=index_path,
                dimension=dimension,
                top_k=top_k,
                metric=metric,
                dtype=dtype,
            )
        elif name == "redis":
            from gptcache.manager.vector_data.redis_vectorstore import RedisVectorStore
            host = kwargs.get("host", "localhost")
            port = kwargs.get("port", "6379")
            user = kwargs.get("user", "")
            password = kwargs.get("password", "")
            namespace = kwargs.get("namespace", "")
            dimension = kwargs.get("dimension", DIMENSION)
            collection_name = kwargs.get("collection_name", COLLECTION_NAME)

            vector_base = RedisVectorStore(
                host=host,
                port=port,
                username=user,
                password=password,
                namespace=namespace,
                dimension=dimension,
                collection_name=collection_name,
                top_k=top_k,
            )
        elif name == "qdrant":
            from gptcache.manager.vector_data.qdrant import QdrantVectorStore
            url = kwargs.get("url", None)
            port = kwargs.get("port", QDRANT_HTTP_PORT)
            grpc_port = kwargs.get("grpc_port", QDRANT_GRPC_PORT)
            prefer_grpc = kwargs.get("prefer_grpc", False)
            https = kwargs.get("https", False)
            api_key = kwargs.get("api_key", None)
            prefix = kwargs.get("prefix", None)
            timeout = kwargs.get("timeout", None)
            host = kwargs.get("host", None)
            collection_name = kwargs.get("collection_name", COLLECTION_NAME)
            location = kwargs.get("location", QDRANT_DEFAULT_LOCATION)
            dimension = kwargs.get("dimension", DIMENSION)
            top_k: int = kwargs.get("top_k", TOP_K)
            flush_interval_sec = kwargs.get("flush_interval_sec", QDRANT_FLUSH_INTERVAL_SEC)
            index_params = kwargs.get("index_params", QDRANT_INDEX_PARAMS)
            vector_base = QdrantVectorStore(
                url=url,
                port=port,
                grpc_port=grpc_port,
                prefer_grpc=prefer_grpc,
                https=https,
                api_key=api_key,
                prefix=prefix,
                timeout=timeout,
                host=host,
                collection_name=collection_name,
                location=location,
                dimension=dimension,
                top_k=top_k,
                flush_interval_sec=flush_interval_sec,
                index_params=index_params,
            )
        elif name == "weaviate":
            from gptcache.manager.vector_data.weaviate import Weaviate

            url = kwargs.get("url", None)
            auth_client_secret = kwargs.get("auth_client_secret", None)
            timeout_config = kwargs.get("timeout_config", WEAVIATE_TIMEOUT_CONFIG)
            proxies = kwargs.get("proxies", None)
            trust_env = kwargs.get("trust_env", False)
            additional_headers = kwargs.get("additional_headers", None)
            startup_period = kwargs.get("startup_period", WEAVIATE_STARTUP_PERIOD)
            embedded_options = kwargs.get("embedded_options", None)
            additional_config = kwargs.get("additional_config", None)
            class_name = kwargs.get("class_name", "GPTCache")
            class_schema = kwargs.get("class_schema", None)

            vector_base = Weaviate(
                url=url,
                auth_client_secret=auth_client_secret,
                timeout_config=timeout_config,
                proxies=proxies,
                trust_env=trust_env,
                additional_headers=additional_headers,
                startup_period=startup_period,
                embedded_options=embedded_options,
                additional_config=additional_config,
                class_name=class_name,
                class_schema=class_schema,
                top_k=top_k,
            )
        else:
            raise NotFoundError("vector store", name)
        return vector_base



class CacheError(Exception):
    """GPTCache base error"""


class NotInitError(CacheError):
    """Raise when the cache has been used before it's inited"""
    def __init__(self):
        super().__init__("The cache should be inited before using")


class NotFoundError(CacheError):
    """Raise when getting an unsupported store."""
    def __init__(self, store_type, current_type_name):
        super().__init__(f"Unsupported ${store_type}: {current_type_name}")


class ParamError(CacheError):
    """Raise when receiving an invalid param."""


class PipInstallError(CacheError):
    """Raise when failed to install package."""
    def __init__(self, package):
        super().__init__(f"Ran into error installing {package}.")


def wrap_error(e: Exception) -> Exception:
    """Add a type to exception `e` while ensuring that the original type is not changed

    Example:
        .. code-block:: python

            import openai

            from gptcache.utils.error import wrap_error


            def raise_error():
                try:
                    raise openai.error.OpenAIError(message="test")
                except openai.error.OpenAIError as e:
                    raise wrap_error(e)


            try:
                raise_error()
            except openai.error.OpenAIError as e:
                print("exception:")
                print(e)

            print("over")
    """
    e.__class__ = type(e.__class__.__name__, (CacheError, e.__class__), {})
    return e



from typing import Tuple, Dict, Any, List

import numpy as np

from gptcache.adapter.api import _get_model
from gptcache.similarity_evaluation import SimilarityEvaluation


def euclidean_distance_calculate(vec_l: np.array, vec_r: np.array):
    return np.sum((vec_l - vec_r) ** 2)


def reweight(weights, length):
    if length >= len(weights):
        return weights
    else:
        reweighted_ws = []
        sum_ws = 0
        for i in range(length):
            sum_ws += weights[i]
        for i in range(length):
            reweighted_ws.append(weights[i] * (1 / sum_ws))
        return reweighted_ws


class SequenceMatchEvaluation(SimilarityEvaluation):
    """
    Evaluate sentence pair similarity using SequenceMatchEvaluation.

    :param weights: List of weights corresponding to each sequence element for calculating the weighted distance.
    :type weights: List[float]
    :param embedding_extractor: The embedding extractor used to obtain embeddings from the text content.
    :type embedding_extractor: gptcache.embedding.base.BaseEmbedding


    Example:
        .. code-block:: python

            from gptcache.similarity_evaluation import SequenceMatchEvaluation
            from gptcache.embedding import Onnx

            weights = [0.5, 0.3, 0.2]
            evaluation = SequenceMatchEvaluation(weights, 'onnx')

            query = {
                'question': 'USER: "foo2" USER: "foo4"',
            }

            cache = {
                'question': 'USER: "foo6" USER: "foo8"',
            }

            score = evaluation.evaluation(query, cache)
    """

    def __init__(
        self, weights: List[float], embedding_extractor: str, embedding_config=None
    ):
        self.embedding_extractor = _get_model(embedding_extractor, embedding_config)
        self.weights = weights

    @staticmethod
    def normalize(vec: np.ndarray):
        """Normalize the input vector.

        :param vec: numpy vector needs to normalize.
        :type vec: numpy.array

        :return: normalized vector.
        """
        magnitude = np.linalg.norm(vec)
        normalized_v = vec / magnitude
        return normalized_v

    def evaluation(
        self, src_dict: Dict[str, Any], cache_dict: Dict[str, Any], **_
    ) -> float:
        """Evaluate the similarity score of pair.

        :param src_dict: the query dictionary to evaluate with cache.
        :type src_dict: Dict
        :param cache_dict: the cache dictionary.
        :type cache_dict: Dict

        :return: evaluation score.
        """
        src_question = src_dict["question"]
        cache_question = cache_dict["question"]
        src_contents = src_question.split("USER: ")
        cache_contents = cache_question.split("USER: ")
        src_contents = [content for content in src_contents if len(content) > 0]
        cache_contents = [content for content in cache_contents if len(content) > 0]
        src_embs = []
        cache_embs = []
        for content in src_contents:
            src_embs.append(
                self.normalize(self.embedding_extractor.to_embeddings(content))
            )
        for content in cache_contents:
            cache_embs.append(
                self.normalize(self.embedding_extractor.to_embeddings(content))
            )
        length = min([len(src_contents), len(cache_contents), len(self.weights)])
        assert length > 0
        ws = self.weights
        ws = ws[::-1]
        ws = reweight(ws, length)
        src_embs = src_embs[::-1]
        cache_embs = cache_embs[::-1]
        weighted_distance = 0
        for i in range(length):
            weighted_distance += (
                4 - euclidean_distance_calculate(src_embs[i], cache_embs[i])
            ) * ws[i]
        return weighted_distance

    def range(self) -> Tuple[float, float]:
        """Range of similarity score.

        :return: minimum and maximum of similarity score.
        """
        return 0.0, 4.0



import datetime
from typing import List, Optional

import numpy as np

from gptcache.manager.scalar_data.base import (
    CacheStorage,
    CacheData,
    Question,
    QuestionDep,
)
from gptcache.utils import import_redis

import_redis()

# pylint: disable=C0413
from redis import Redis
from redis.client import Pipeline
from redis_om import get_redis_connection
from redis_om import JsonModel, EmbeddedJsonModel, NotFoundError, Field, Migrator


def get_models(global_key: str, redis_connection: Redis):
    """
    Get all the models for the given global key and redis connection.
    :param global_key: Global key will be used as a prefix for all the keys
    :type global_key: str

    :param redis_connection: Redis connection to use for all the models.
    Note: This needs to be explicitly mentioned in `Meta` class for each Object Model,
    otherwise it will use the default connection from the pool.
    :type redis_connection: Redis
    """

    class Counter:
        """
        counter collection
        """
        key_name = global_key + ":counter"
        database = redis_connection

        @classmethod
        def incr(cls):
            cls.database.incr(cls.key_name)

        @classmethod
        def get(cls):
            return cls.database.get(cls.key_name)

    class EmbeddingType:
        """
        Directly using bytes for embedding data is not supported by redis-om as of now.
        Custom type for embedding data. This will be stored as bytes in redis.
        Latin-1 encoding is used to convert the bytes to string and vice versa.
        """

        def __init__(self, data: bytes):
            self.data = data

        @classmethod
        def __get_validators__(cls):
            yield cls.validate

        @classmethod
        def validate(cls, v: [np.array, bytes]):
            if isinstance(v, np.ndarray):
                return cls(v.astype(np.float32).tobytes())
            elif isinstance(v, bytes):
                return cls(v)

            return cls(v)

        def to_numpy(self) -> np.ndarray:
            return np.frombuffer(self.data.encode("latin-1"), dtype=np.float32)

        def __repr__(self):
            return f"{self.data}"

    class Answers(EmbeddedJsonModel):
        """
        answer collection
        """

        answer: str
        answer_type: int

        class Meta:
            database = redis_connection

    class QuestionDeps(EmbeddedJsonModel):
        """
        Question Dep collection
        """

        dep_name: str
        dep_data: str
        dep_type: int

    class Questions(JsonModel):
        """
        questions collection
        """

        question: str = Field(index=True)
        create_on: datetime.datetime
        last_access: datetime.datetime
        deleted: int = Field(index=True)
        answers: List[Answers]
        deps: List[QuestionDeps]
        embedding: EmbeddingType

        class Meta:
            global_key_prefix = global_key
            model_key_prefix = "questions"
            database = redis_connection

        class Config:
            json_encoders = {
                EmbeddingType: lambda n: n.data.decode("latin-1")
                if isinstance(n.data, bytes) else n.data
            }

    class Sessions(JsonModel):
        """
        session collection
        """

        class Meta:
            global_key_prefix = global_key
            model_key_prefix = "sessions"
            database = redis_connection

        session_id: str = Field(index=True)
        session_question: str
        question_id: str = Field(index=True)

    class Report(JsonModel):
        """
        Report collection
        """

        class Meta:
            global_key_prefix = global_key
            model_key_prefix = "report"
            database = redis_connection

        user_question: str
        cache_question_id: int = Field(index=True)
        cache_question: str
        cache_answer: str
        similarity: float = Field(index=True)
        cache_delta_time: float = Field(index=True)
        cache_time: datetime.datetime = Field(index=True)
        extra: Optional[str]

    return Questions, Answers, QuestionDeps, Sessions, Counter, Report


class RedisCacheStorage(CacheStorage):
    """
     Using redis-om as OM to store data in redis cache storage

    :param host: redis host, default value 'localhost'
    :type host: str
    :param port: redis port, default value 27017
    :type port: int
    :param global_key_prefix: A global prefix for keys against which data is stored.
    For example, for a global_key_prefix ='gptcache', keys would be constructed would look like this:
    gptcache:questions:abc123
    :type global_key_prefix: str
    :param maxmemory: Maximum memory to use for redis cache storage
    :type maxmemory: str
    :param policy: Policy to use for eviction, default value 'allkeys-lru'
    :type policy: str
    :param ttl: Time to live for keys in milliseconds, default value None
    :type ttl: int
    :param maxmemory_samples: Number of keys to sample when evicting keys
    :type maxmemory_samples: int
    :param kwargs: Additional parameters to provide in order to create redis om connection
    Example:
        .. code-block:: python

            from gptcache.manager import CacheBase, manager_factory

            cache_store = CacheBase('redis',
                redis_host="localhost",
                redis_port=6379,
                global_key_prefix="gptcache",
            )
            # or
            data_manager = manager_factory("mongo,faiss", data_dir="./workspace",
                scalar_params={
                    "redis_host"="localhost",
                    "redis_port"=6379,
                    "global_key_prefix"="gptcache",
                },
                vector_params={"dimension": 128},
            )
    """

    def __init__(
            self,
            global_key_prefix: str = "gptcache",
            host: str = "localhost",
            port: int = 6379,
            maxmemory: str = None,
            policy: str = None,
            ttl: int = None,
            maxmemory_samples: int = None,
            **kwargs
    ):
        self.con = get_redis_connection(host=host, port=port, **kwargs)
        self.default_ttl = ttl
        self.init_eviction_params(policy=policy, maxmemory=maxmemory, maxmemory_samples=maxmemory_samples, ttl=ttl)
        (
            self._ques,
            self._answer,
            self._ques_dep,
            self._session,
            self._counter,
            self._report,
        ) = get_models(global_key_prefix, redis_connection=self.con)

        Migrator().run()

    def init_eviction_params(self, policy, maxmemory, maxmemory_samples, ttl):
        self.default_ttl = ttl
        if maxmemory:
            self.con.config_set("maxmemory", maxmemory)
        if policy:
            self.con.config_set("maxmemory-policy", policy)
        if maxmemory_samples:
            self.con.config_set("maxmemory-samples", maxmemory_samples)

    def create(self):
        pass

    def _insert(self, data: CacheData, pipeline: Pipeline = None):
        self._counter.incr()
        pk = str(self._counter.get())
        answers = data.answers if isinstance(data.answers, list) else [data.answers]
        all_data = []
        for answer in answers:
            answer_data = self._answer(
                answer=answer.answer,
                answer_type=int(answer.answer_type),
            )
            all_data.append(answer_data)

        all_deps = []
        if isinstance(data.question, Question) and data.question.deps is not None:
            for dep in data.question.deps:
                all_deps.append(
                    self._ques_dep(
                        dep_name=dep.name,
                        dep_data=dep.data,
                        dep_type=dep.dep_type,
                    )
                )
        embedding_data = (
            data.embedding_data
            if data.embedding_data is not None
            else None
        )
        ques_data = self._ques(
            pk=pk,
            question=data.question
            if isinstance(data.question, str)
            else data.question.content,
            create_on=datetime.datetime.utcnow(),
            last_access=datetime.datetime.utcnow(),
            deleted=0,
            answers=answers,
            deps=all_deps,
            embedding=embedding_data
        )

        ques_data.save(pipeline)

        if data.session_id:
            session_data = self._session(
                question_id=ques_data.pk,
                session_id=data.session_id,
                session_question=data.question
                if isinstance(data.question, str)
                else data.question.content,
            )
            session_data.save(pipeline)
        if self.default_ttl:
            ques_data.expire(self.default_ttl, pipeline=pipeline)
        return int(ques_data.pk)

    def batch_insert(self, all_data: List[CacheData]):
        ids = []
        with self.con.pipeline() as pipeline:
            for data in all_data:
                ids.append(self._insert(data, pipeline=pipeline))
            pipeline.execute()
        return ids

    def get_data_by_id(self, key: str):
        key = str(key)
        try:
            qs = self._ques.get(pk=key)
        except NotFoundError:
            return None

        qs.update(last_access=datetime.datetime.utcnow())
        res_ans = [(item.answer, item.answer_type) for item in qs.answers]
        res_deps = [
            QuestionDep(item.dep_name, item.dep_data, item.dep_type) for item in qs.deps
        ]

        session_ids = [
            obj.session_id
            for obj in self._session.find(self._session.question_id == key).all()
        ]
        if self.default_ttl:
            qs.expire(self.default_ttl)
        return CacheData(
            question=qs.question if not res_deps else Question(qs.question, res_deps),
            answers=res_ans,
            embedding_data=qs.embedding.to_numpy(),
            session_id=session_ids,
            create_on=qs.create_on,
            last_access=qs.last_access,
        )

    def mark_deleted(self, keys):
        result = self._ques.find(self._ques.pk << keys).all()
        for qs in result:
            qs.update(deleted=-1)

    def clear_deleted_data(self):
        with self.con.pipeline() as pipeline:
            qs_to_delete = self._ques.find(self._ques.deleted == -1).all()
            self._ques.delete_many(qs_to_delete, pipeline)

            q_ids = [qs.pk for qs in qs_to_delete]
            sessions_to_delete = self._session.find(
                self._session.question_id << q_ids
            ).all()
            self._session.delete_many(sessions_to_delete, pipeline)

            pipeline.execute()

    def get_ids(self, deleted=True):
        state = -1 if deleted else 0
        res = [
            int(obj.pk) for obj in self._ques.find(self._ques.deleted == state).all()
        ]
        return res

    def count(self, state: int = 0, is_all: bool = False):
        if is_all:
            return self._ques.find().count()
        return self._ques.find(self._ques.deleted == state).count()

    def add_session(self, question_id, session_id, session_question):
        self._session(
            question_id=question_id,
            session_id=session_id,
            session_question=session_question,
        ).save()

    def list_sessions(self, session_id=None, key=None):
        if session_id and key:
            self._session.find(
                self._session.session_id == session_id
                and self._session.question_id == key
            ).all()
        if key:
            key = str(key)
            return self._session.find(self._session.question_id == key).all()
        if session_id:
            return self._session.find(self._session.session_id == session_id).all()
        return self._session.find().all()

    def delete_session(self, keys: List[str]):
        keys = [str(key) for key in keys]
        with self.con.pipeline() as pipeline:
            sessions_to_delete = self._session.find(
                self._session.question_id << keys
            ).all()
            self._session.delete_many(sessions_to_delete, pipeline)
            pipeline.execute()

    def report_cache(self, user_question, cache_question, cache_question_id, cache_answer, similarity_value,
                     cache_delta_time):
        self._report(
            user_question=user_question,
            cache_question=cache_question,
            cache_question_id=cache_question_id,
            cache_answer=cache_answer,
            similarity=similarity_value,
            cache_delta_time=cache_delta_time,
            cache_time=datetime.datetime.utcnow(),
        ).save()

    def close(self):
        self.con.close()



# pylint: disable=wrong-import-position
from typing import Any, Optional, Callable

import gptcache.processor.post
import gptcache.processor.pre
from gptcache import Cache, cache, Config
from gptcache.adapter.adapter import adapt
from gptcache.embedding import (
    Onnx,
    Huggingface,
    SBERT,
    FastText,
    Data2VecAudio,
    Timm,
    ViT,
    OpenAI,
    Cohere,
    Rwkv,
    PaddleNLP,
    UForm,
)
from gptcache.embedding.base import BaseEmbedding
from gptcache.manager import manager_factory
from gptcache.manager.data_manager import DataManager
from gptcache.processor.context import (
    SummarizationContextProcess,
    SelectiveContextProcess,
    ConcatContextProcess,
)
from gptcache.processor.post import temperature_softmax
from gptcache.processor.pre import get_prompt
from gptcache.similarity_evaluation import (
    SearchDistanceEvaluation,
    NumpyNormEvaluation,
    OnnxModelEvaluation,
    ExactMatchEvaluation,
    KReciprocalEvaluation,
    SimilarityEvaluation,
    CohereRerankEvaluation,
    SequenceMatchEvaluation,
    TimeEvaluation,
    SbertCrossencoderEvaluation
)
from gptcache.utils import import_ruamel


def _cache_data_converter(cache_data):
    """For cache results, do nothing"""
    return cache_data


def _update_cache_callback_none(
    llm_data, update_cache_func, *args, **kwargs  # pylint: disable=W0613
) -> None:
    """When updating cached data, do nothing, because currently only cached queries are processed"""
    return None


def _llm_handle_none(*llm_args, **llm_kwargs) -> None:  # pylint: disable=W0613
    """Do nothing on a cache miss"""
    return None


def _update_cache_callback(
    llm_data, update_cache_func, *args, **kwargs
):  # pylint: disable=W0613
    """Save the `llm_data` to cache storage"""
    update_cache_func(llm_data)


def put(prompt: str, data: Any, **kwargs) -> None:
    """put api, put qa pair information to GPTCache
    Please make sure that the `pre_embedding_func` param is `get_prompt` when initializing the cache

    :param prompt: the cache data key, usually question text
    :type prompt: str
    :param data: the cache data value, usually answer text
    :type data: Any
    :param kwargs: list of user-defined parameters
    :type kwargs: Dict

    Example:
        .. code-block:: python

            from gptcache.adapter.api import put
            from gptcache.processor.pre import get_prompt

            cache.init(pre_embedding_func=get_prompt)
            put("hello", "foo")
    """

    def llm_handle(*llm_args, **llm_kwargs):  # pylint: disable=W0613
        return data

    adapt(
        llm_handle,
        _cache_data_converter,
        _update_cache_callback,
        cache_skip=True,
        prompt=prompt,
        **kwargs,
    )


def get(prompt: str, **kwargs) -> Any:
    """get api, get the cache data according to the `prompt`
    Please make sure that the `pre_embedding_func` param is `get_prompt` when initializing the cache

    :param prompt: the cache data key, usually question text
    :type prompt: str
    :param kwargs: list of user-defined parameters
    :type kwargs: Dict

    Example:
        .. code-block:: python

            from gptcache.adapter.api import put, get
            from gptcache.processor.pre import get_prompt

            cache.init(pre_embedding_func=get_prompt)
            put("hello", "foo")
            print(get("hello"))
    """
    res = adapt(
        _llm_handle_none,
        _cache_data_converter,
        _update_cache_callback_none,
        prompt=prompt,
        **kwargs,
    )
    return res


def init_similar_cache(
    data_dir: str = "api_cache",
    cache_obj: Optional[Cache] = None,
    pre_func: Callable = get_prompt,
    embedding: Optional[BaseEmbedding] = None,
    data_manager: Optional[DataManager] = None,
    evaluation: Optional[SimilarityEvaluation] = None,
    post_func: Callable = temperature_softmax,
    config: Config = Config(),
):
    """Provide a quick way to initialize cache for api service

    :param data_dir: cache data storage directory
    :type data_dir: str
    :param cache_obj: specify to initialize the Cache object, if not specified, initialize the global object
    :type cache_obj: Optional[Cache]
    :param pre_func: pre-processing of the cache input text
    :type pre_func: Callable
    :param embedding: embedding object
    :type embedding: BaseEmbedding
    :param data_manager: data manager object
    :type data_manager: DataManager
    :param evaluation: similarity evaluation object
    :type evaluation: SimilarityEvaluation
    :param post_func: post-processing of the cached result list, the most similar result is taken by default
    :type post_func: Callable[[List[Any]], Any]
    :param config: cache configuration, the core is similar threshold
    :type config: Config
    :return: None

    Example:
        .. code-block:: python

            from gptcache.adapter.api import put, get, init_similar_cache

            init_similar_cache()
            put("hello", "foo")
            print(get("hello"))
    """
    if not embedding:
        embedding = Onnx()
    if not data_manager:
        data_manager = manager_factory(
            "sqlite,faiss",
            data_dir=data_dir,
            vector_params={"dimension": embedding.dimension},
        )
    if not evaluation:
        evaluation = SearchDistanceEvaluation()
    cache_obj = cache_obj if cache_obj else cache
    cache_obj.init(
        pre_embedding_func=pre_func,
        embedding_func=embedding.to_embeddings,
        data_manager=data_manager,
        similarity_evaluation=evaluation,
        post_process_messages_func=post_func,
        config=config,
    )


def init_similar_cache_from_config(config_dir: str, cache_obj: Optional[Cache] = None):
    import_ruamel()
    from ruamel.yaml import YAML  # pylint: disable=C0415

    if config_dir:
        with open(config_dir, "r", encoding="utf-8") as f:
            yaml = YAML(typ="unsafe", pure=True)
            init_conf = yaml.load(f)
    else:
        init_conf = {}

    # Due to the problem with the first naming, it is reserved to ensure compatibility
    embedding = init_conf.get("model_source", "")
    if not embedding:
        embedding = init_conf.get("embedding", "onnx")
    # ditto
    embedding_config = init_conf.get("model_config", {})
    if not embedding_config:
        embedding_config = init_conf.get("embedding_config", {})
    embedding_model = _get_model(embedding, embedding_config)

    storage_config = init_conf.get("storage_config", {})
    storage_config.setdefault("manager", "sqlite,faiss")
    storage_config.setdefault("data_dir", "gptcache_data")
    storage_config.setdefault("vector_params", {})
    storage_config["vector_params"] = storage_config["vector_params"] or {}
    storage_config["vector_params"]["dimension"] = embedding_model.dimension
    data_manager = manager_factory(**storage_config)

    eval_strategy = init_conf.get("evaluation", "distance")
    # Due to the problem with the first naming, it is reserved to ensure compatibility
    eval_config = init_conf.get("evaluation_kws", {})
    if not eval_config:
        eval_config = init_conf.get("evaluation_config", {})
    evaluation = _get_eval(eval_strategy, eval_config)

    cache_obj = cache_obj if cache_obj else cache

    pre_process = init_conf.get("pre_context_function")
    if pre_process:
        pre_func = _get_pre_context_function(
            pre_process, init_conf.get("pre_context_config")
        )
        pre_func = pre_func.pre_process
    else:
        pre_process = init_conf.get("pre_function", "get_prompt")
        pre_func = _get_pre_func(pre_process)

    post_process = init_conf.get("post_function", "first")
    post_func = _get_post_func(post_process)

    config_kws = init_conf.get("config", {}) or {}
    config = Config(**config_kws)

    cache_obj.init(
        pre_embedding_func=pre_func,
        embedding_func=embedding_model.to_embeddings,
        data_manager=data_manager,
        similarity_evaluation=evaluation,
        post_process_messages_func=post_func,
        config=config,
    )

    return init_conf


def _get_model(model_src, model_config=None):
    model_src = model_src.lower()
    model_config = model_config or {}

    if model_src == "onnx":
        return Onnx(**model_config)
    if model_src == "huggingface":
        return Huggingface(**model_config)
    if model_src == "sbert":
        return SBERT(**model_config)
    if model_src == "fasttext":
        return FastText(**model_config)
    if model_src == "data2vecaudio":
        return Data2VecAudio(**model_config)
    if model_src == "timm":
        return Timm(**model_config)
    if model_src == "vit":
        return ViT(**model_config)
    if model_src == "openai":
        return OpenAI(**model_config)
    if model_src == "cohere":
        return Cohere(**model_config)
    if model_src == "rwkv":
        return Rwkv(**model_config)
    if model_src == "paddlenlp":
        return PaddleNLP(**model_config)
    if model_src == "uform":
        return UForm(**model_config)


def _get_eval(strategy, kws=None):
    strategy = strategy.lower()
    kws = kws or {}

    if "distance" in strategy:
        return SearchDistanceEvaluation(**kws)
    if "np" in strategy:
        return NumpyNormEvaluation(**kws)
    if "exact" in strategy:
        return ExactMatchEvaluation()
    if "onnx" in strategy:
        return OnnxModelEvaluation(**kws)
    if "kreciprocal" in strategy:
        return KReciprocalEvaluation(**kws)
    if "cohere" in strategy:
        return CohereRerankEvaluation(**kws)
    if "sequence_match" in strategy:
        return SequenceMatchEvaluation(**kws)
    if "time" in strategy:
        return TimeEvaluation(**kws)
    if "sbert_crossencoder" in strategy:
        return SbertCrossencoderEvaluation(**kws)


def _get_pre_func(pre_process):
    return getattr(gptcache.processor.pre, pre_process)


def _get_pre_context_function(pre_context_process, kws=None):
    pre_context_process = pre_context_process.lower()
    kws = kws or {}
    if pre_context_process in "summarization":
        return SummarizationContextProcess(**kws)
    if pre_context_process in "selective":
        return SelectiveContextProcess(**kws)
    if pre_context_process in "concat":
        return ConcatContextProcess()


def _get_post_func(post_process):
    return getattr(gptcache.processor.post, post_process)



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum
from typing import Optional, Any, List, Union, Dict

import numpy as np


class DataType(IntEnum):
    STR = 0
    IMAGE_BASE64 = 1
    IMAGE_URL = 2


@dataclass
class Answer:
    """
    data_type:
        0: str
        1: base64 image
    """

    answer: Any
    answer_type: int = DataType.STR


@dataclass
class QuestionDep:
    """
    QuestionDep
    """

    name: str
    data: str
    dep_type: int = DataType.STR

    @classmethod
    def from_dict(cls, d: Dict):
        return cls(name=d["name"], data=d["data"], dep_type=d["dep_type"])


@dataclass
class Question:
    """
    Question
    """

    content: str
    deps: Optional[List[QuestionDep]] = None

    @classmethod
    def from_dict(cls, d: Dict):
        deps = []
        for dep in d["deps"]:
            deps.append(QuestionDep.from_dict(dep))
        return cls(d["content"], deps)


@dataclass
class CacheData:
    """
    CacheData
    """

    question: Union[str, Question]
    answers: List[Answer]
    embedding_data: Optional[np.ndarray] = None
    session_id: Optional[str] = None
    create_on: Optional[datetime] = None
    last_access: Optional[datetime] = None

    def __init__(
        self,
        question,
        answers,
        embedding_data=None,
        session_id=None,
        create_on=None,
        last_access=None,
    ):
        self.question = question
        self.answers = []
        if isinstance(answers, (str, Answer)):
            answers = [answers]
        for data in answers:
            if isinstance(data, (list, tuple)):
                self.answers.append(Answer(*data))
            elif isinstance(data, Answer):
                self.answers.append(data)
            else:
                self.answers.append(Answer(answer=data))
        self.embedding_data = embedding_data
        self.session_id = session_id
        self.create_on = create_on
        self.last_access = last_access


class CacheStorage(metaclass=ABCMeta):
    """
    BaseStorage for scalar data.
    """

    @abstractmethod
    def create(self):
        pass

    @abstractmethod
    def batch_insert(self, all_data: List[CacheData]):
        pass

    @abstractmethod
    def get_data_by_id(self, key):
        pass

    @abstractmethod
    def mark_deleted(self, keys):
        pass

    @abstractmethod
    def clear_deleted_data(self):
        pass

    @abstractmethod
    def get_ids(self, deleted=True):
        pass

    @abstractmethod
    def count(self, state: int = 0, is_all: bool = False):
        pass

    def flush(self):
        pass

    @abstractmethod
    def add_session(self, question_id, session_id, session_question):
        pass

    @abstractmethod
    def list_sessions(self, session_id, key):
        pass

    @abstractmethod
    def delete_session(self, keys):
        pass

    @abstractmethod
    def report_cache(
        self,
        user_question,
        cache_question,
        cache_question_id,
        cache_answer,
        similarity_value,
        cache_delta_time,
    ):
        pass

    @abstractmethod
    def close(self):
        pass



import os
from typing import List

import numpy as np

from gptcache.manager.vector_data.base import VectorBase, VectorData
from gptcache.utils import import_usearch

import_usearch()

from usearch.index import Index  # pylint: disable=C0413
from usearch.compiled import MetricKind  # pylint: disable=C0413


class USearch(VectorBase):
    """vector store: Usearch

    :param index_path: the path to Usearch index, defaults to 'index.usearch'.
    :type index_path: str
    :param dimension: the dimension of the vector, defaults to 0.
    :type dimension: int
    :param top_k: the number of the vectors results to return, defaults to 1.
    :type top_k: int
    :param metric: the distance mrtric. 'l2', 'haversine' or other, default = 'ip'
    :type metric: str
    :param dtype: the quantization dtype, 'f16' or 'f8' if needed, default = 'f32'
    :type dtype: str
    :param connectivity: the frequency of the connections in the graph, optional
    :type connectivity: int
    :param expansion_add: the recall of indexing, optional
    :type expansion_add: int
    :param expansion_search: the quality of search, optional
    :type expansion_search: int
    """

    def __init__(
        self,
        index_file_path: str = "index.usearch",
        dimension: int = 64,
        top_k: int = 1,
        metric: str = "cos",
        dtype: str = "f32",
        connectivity: int = 16,
        expansion_add: int = 128,
        expansion_search: int = 64,
    ):
        self._index_file_path = index_file_path
        self._dimension = dimension
        self._top_k = top_k
        self._index = Index(
            ndim=self._dimension,
            metric=getattr(MetricKind, metric.lower().capitalize()),
            dtype=dtype,
            connectivity=connectivity,
            expansion_add=expansion_add,
            expansion_search=expansion_search,
        )
        if os.path.isfile(self._index_file_path):
            self._index.load(self._index_file_path)

    def mul_add(self, datas: List[VectorData]):
        data_array, id_array = map(list, zip(*((data.data, data.id) for data in datas)))
        np_data = np.array(data_array).astype("float32")
        ids = np.array(id_array, dtype=np.longlong)
        self._index.add(ids, np_data)

    def search(self, data: np.ndarray, top_k: int = -1):
        if top_k == -1:
            top_k = self._top_k
        np_data = np.array(data).astype("float32").reshape(1, -1)
        ids, dist, _ = self._index.search(np_data, top_k)
        return list(zip(dist[0], ids[0]))

    def rebuild(self, ids=None):
        return True

    def delete(self, ids):
        raise NotImplementedError

    def flush(self):
        self._index.save(self._index_file_path)

    def close(self):
        self.flush()

    def count(self):
        return len(self._index)



import numpy as np


def softmax(x: list):
    x = np.array(x)
    assert len(x.shape) == 1, f"Expect to get a shape of (len,) but got {x.shape}, x value: {x}."
    max_val = x.max()
    e_x = np.exp(x - max_val)
    return e_x / e_x.sum()



import base64
import json
import os
import time
from io import BytesIO
from typing import Any, AsyncGenerator, Iterator, List

from gptcache import cache
from gptcache.adapter.adapter import aadapt, adapt
from gptcache.adapter.base import BaseCacheLLM
from gptcache.manager.scalar_data.base import Answer, DataType
from gptcache.utils import import_openai, import_pillow
from gptcache.utils.error import wrap_error
from gptcache.utils.response import (
    get_audio_text_from_openai_answer,
    get_image_from_openai_b64,
    get_image_from_openai_url,
    get_message_from_openai_answer,
    get_stream_message_from_openai_answer,
    get_text_from_openai_answer,
)
from gptcache.utils.token import token_counter

import_openai()

# pylint: disable=C0413
# pylint: disable=E1102
import openai


class ChatCompletion(openai.ChatCompletion, BaseCacheLLM):
    """Openai ChatCompletion Wrapper

    Example:
        .. code-block:: python

            from gptcache import cache
            from gptcache.processor.pre import get_prompt
            # init gptcache
            cache.init()
            cache.set_openai_key()

            from gptcache.adapter import openai
            # run ChatCompletion model with gptcache
            response = openai.ChatCompletion.create(
                          model='gpt-3.5-turbo',
                          messages=[
                            {
                                'role': 'user',
                                'content': "what's github"
                            }],
                        )
            response_content = response['choices'][0]['message']['content']
    """

    @classmethod
    def _llm_handler(cls, *llm_args, **llm_kwargs):
        try:
            return (
                super().create(*llm_args, **llm_kwargs)
                if cls.llm is None
                else cls.llm(*llm_args, **llm_kwargs)
            )
        except openai.OpenAIError as e:
            raise wrap_error(e) from e

    @classmethod
    async def _allm_handler(cls, *llm_args, **llm_kwargs):
        try:
            return (
                (await super().acreate(*llm_args, **llm_kwargs))
                if cls.llm is None
                else await cls.llm(*llm_args, **llm_kwargs)
            )
        except openai.OpenAIError as e:
            raise wrap_error(e) from e

    @staticmethod
    def _update_cache_callback(
        llm_data, update_cache_func, *args, **kwargs
    ):  # pylint: disable=unused-argument
        if isinstance(llm_data, AsyncGenerator):

            async def hook_openai_data(it):
                total_answer = ""
                async for item in it:
                    total_answer += get_stream_message_from_openai_answer(item)
                    yield item
                update_cache_func(Answer(total_answer, DataType.STR))

            return hook_openai_data(llm_data)
        elif not isinstance(llm_data, Iterator):
            update_cache_func(
                Answer(get_message_from_openai_answer(llm_data), DataType.STR)
            )
            return llm_data
        else:

            def hook_openai_data(it):
                total_answer = ""
                for item in it:
                    total_answer += get_stream_message_from_openai_answer(item)
                    yield item
                update_cache_func(Answer(total_answer, DataType.STR))

            return hook_openai_data(llm_data)

    @classmethod
    def create(cls, *args, **kwargs):
        chat_cache = kwargs.get("cache_obj", cache)
        enable_token_counter = chat_cache.config.enable_token_counter

        def cache_data_convert(cache_data):
            if enable_token_counter:
                input_token = _num_tokens_from_messages(kwargs.get("messages"))
                output_token = token_counter(cache_data)
                saved_token = [input_token, output_token]
            else:
                saved_token = [0, 0]
            if kwargs.get("stream", False):
                return _construct_stream_resp_from_cache(cache_data, saved_token)
            return _construct_resp_from_cache(cache_data, saved_token)

        kwargs = cls.fill_base_args(**kwargs)
        return adapt(
            cls._llm_handler,
            cache_data_convert,
            cls._update_cache_callback,
            *args,
            **kwargs,
        )

    @classmethod
    async def acreate(cls, *args, **kwargs):
        chat_cache = kwargs.get("cache_obj", cache)
        enable_token_counter = chat_cache.config.enable_token_counter

        def cache_data_convert(cache_data):
            if enable_token_counter:
                input_token = _num_tokens_from_messages(kwargs.get("messages"))
                output_token = token_counter(cache_data)
                saved_token = [input_token, output_token]
            else:
                saved_token = [0, 0]
            if kwargs.get("stream", False):
                return async_iter(
                    _construct_stream_resp_from_cache(cache_data, saved_token)
                )
            return _construct_resp_from_cache(cache_data, saved_token)

        kwargs = cls.fill_base_args(**kwargs)
        return await aadapt(
            cls._allm_handler,
            cache_data_convert,
            cls._update_cache_callback,
            *args,
            **kwargs,
        )


async def async_iter(input_list):
    for item in input_list:
        yield item


class Completion(openai.Completion, BaseCacheLLM):
    """Openai Completion Wrapper

    Example:
        .. code-block:: python

            from gptcache import cache
            from gptcache.processor.pre import get_prompt
            # init gptcache
            cache.init()
            cache.set_openai_key()

            from gptcache.adapter import openai
            # run Completion model with gptcache
            response = openai.Completion.create(model="text-davinci-003",
                                                prompt="Hello world.")
            response_text = response["choices"][0]["text"]
    """

    @classmethod
    def _llm_handler(cls, *llm_args, **llm_kwargs):
        try:
            return (
                super().create(*llm_args, **llm_kwargs)
                if not cls.llm
                else cls.llm(*llm_args, **llm_kwargs)
            )
        except openai.OpenAIError as e:
            raise wrap_error(e) from e

    @classmethod
    async def _allm_handler(cls, *llm_args, **llm_kwargs):
        try:
            return (
                (await super().acreate(*llm_args, **llm_kwargs))
                if cls.llm is None
                else await cls.llm(*llm_args, **llm_kwargs)
            )
        except openai.OpenAIError as e:
            raise wrap_error(e) from e

    @staticmethod
    def _cache_data_convert(cache_data):
        return _construct_text_from_cache(cache_data)

    @staticmethod
    def _update_cache_callback(
        llm_data, update_cache_func, *args, **kwargs
    ):  # pylint: disable=unused-argument
        update_cache_func(Answer(get_text_from_openai_answer(llm_data), DataType.STR))
        return llm_data

    @classmethod
    def create(cls, *args, **kwargs):
        kwargs = cls.fill_base_args(**kwargs)
        return adapt(
            cls._llm_handler,
            cls._cache_data_convert,
            cls._update_cache_callback,
            *args,
            **kwargs,
        )

    @classmethod
    async def acreate(cls, *args, **kwargs):
        kwargs = cls.fill_base_args(**kwargs)
        return await aadapt(
            cls._allm_handler,
            cls._cache_data_convert,
            cls._update_cache_callback,
            *args,
            **kwargs,
        )


class Audio(openai.Audio):
    """Openai Audio Wrapper

    Example:
        .. code-block:: python

            from gptcache import cache
            from gptcache.processor.pre import get_file_bytes
            # init gptcache
            cache.init(pre_embedding_func=get_file_bytes)
            cache.set_openai_key()

            from gptcache.adapter import openai
            # run audio transcribe model with gptcache
            audio_file= open("/path/to/audio.mp3", "rb")
            transcript = openai.Audio.transcribe("whisper-1", audio_file)

            # run audio transcribe model with gptcache
            audio_file= open("/path/to/audio.mp3", "rb")
            transcript = openai.Audio.translate("whisper-1", audio_file)
    """

    @classmethod
    def transcribe(cls, model: str, file: Any, *args, **kwargs):
        def llm_handler(*llm_args, **llm_kwargs):
            try:
                return super(Audio, cls).transcribe(*llm_args, **llm_kwargs)
            except openai.OpenAIError as e:
                raise wrap_error(e) from e

        def cache_data_convert(cache_data):
            return _construct_audio_text_from_cache(cache_data)

        def update_cache_callback(
            llm_data, update_cache_func, *args, **kwargs
        ):  # pylint: disable=unused-argument
            update_cache_func(
                Answer(get_audio_text_from_openai_answer(llm_data), DataType.STR)
            )
            return llm_data

        return adapt(
            llm_handler,
            cache_data_convert,
            update_cache_callback,
            model=model,
            file=file,
            *args,
            **kwargs,
        )

    @classmethod
    def translate(cls, model: str, file: Any, *args, **kwargs):
        def llm_handler(*llm_args, **llm_kwargs):
            try:
                return super(Audio, cls).translate(*llm_args, **llm_kwargs)
            except openai.OpenAIError as e:
                raise wrap_error(e) from e

        def cache_data_convert(cache_data):
            return _construct_audio_text_from_cache(cache_data)

        def update_cache_callback(
            llm_data, update_cache_func, *args, **kwargs
        ):  # pylint: disable=unused-argument
            update_cache_func(
                Answer(get_audio_text_from_openai_answer(llm_data), DataType.STR)
            )
            return llm_data

        return adapt(
            llm_handler,
            cache_data_convert,
            update_cache_callback,
            model=model,
            file=file,
            *args,
            **kwargs,
        )


class Image(openai.Image):
    """Openai Image Wrapper

    Example:
        .. code-block:: python

            from gptcache import cache
            from gptcache.processor.pre import get_prompt
            # init gptcache
            cache.init(pre_embedding_func=get_prompt)
            cache.set_openai_key()

            from gptcache.adapter import openai
            # run image generation model with gptcache
            response = openai.Image.create(
              prompt="a white siamese cat",
              n=1,
              size="256x256"
            )
            response_url = response['data'][0]['url']
    """

    @classmethod
    def _llm_handler(cls, *llm_args, **llm_kwargs):
        try:
            return super().create(*llm_args, **llm_kwargs)
        except openai.OpenAIError as e:
            raise wrap_error(e) from e

    @classmethod
    def create(cls, *args, **kwargs):
        response_format = kwargs.pop("response_format", "url")
        size = kwargs.pop("size", "256x256")

        def cache_data_convert(cache_data):
            return _construct_image_create_resp_from_cache(
                image_data=cache_data, response_format=response_format, size=size
            )

        def update_cache_callback(
            llm_data, update_cache_func, *args, **kwargs
        ):  # pylint: disable=unused-argument
            if response_format == "b64_json":
                img_b64 = get_image_from_openai_b64(llm_data)
                if isinstance(img_b64, str):
                    img_b64 = img_b64.encode("ascii")
                update_cache_func(Answer(img_b64, DataType.IMAGE_BASE64))
            elif response_format == "url":
                update_cache_func(
                    Answer(get_image_from_openai_url(llm_data), DataType.IMAGE_URL)
                )
            return llm_data

        return adapt(
            cls._llm_handler,
            cache_data_convert,
            update_cache_callback,
            response_format=response_format,
            size=size,
            *args,
            **kwargs,
        )


class Moderation(openai.Moderation, BaseCacheLLM):
    """Openai Moderation Wrapper

    Example:
        .. code-block:: python

            from gptcache.adapter import openai
            from gptcache.adapter.api import init_similar_cache
            from gptcache.processor.pre import get_openai_moderation_input

            init_similar_cache(pre_func=get_openai_moderation_input)
            openai.Moderation.create(
                input="I want to kill them.",
            )
    """

    @classmethod
    def _llm_handler(cls, *llm_args, **llm_kwargs):
        try:
            return (
                super().create(*llm_args, **llm_kwargs)
                if not cls.llm
                else cls.llm(*llm_args, **llm_kwargs)
            )
        except openai.OpenAIError as e:
            raise wrap_error(e) from e

    @classmethod
    def _cache_data_convert(cls, cache_data):
        return json.loads(cache_data)

    @classmethod
    def _update_cache_callback(
        cls, llm_data, update_cache_func, *args, **kwargs
    ):  # pylint: disable=unused-argument
        update_cache_func(Answer(json.dumps(llm_data, indent=4), DataType.STR))
        return llm_data

    @classmethod
    def create(cls, *args, **kwargs):
        kwargs = cls.fill_base_args(**kwargs)
        res = adapt(
            cls._llm_handler,
            cls._cache_data_convert,
            cls._update_cache_callback,
            *args,
            **kwargs,
        )

        input_request_param = kwargs.get("input")
        expect_res_len = 1
        if isinstance(input_request_param, List):
            expect_res_len = len(input_request_param)
        if len(res.get("results")) != expect_res_len:
            kwargs["cache_skip"] = True
            res = adapt(
                cls._llm_handler,
                cls._cache_data_convert,
                cls._update_cache_callback,
                *args,
                **kwargs,
            )
        return res


def _construct_resp_from_cache(return_message, saved_token):
    return {
        "gptcache": True,
        "saved_token": saved_token,
        "choices": [
            {
                "message": {"role": "assistant", "content": return_message},
                "finish_reason": "stop",
                "index": 0,
            }
        ],
        "created": int(time.time()),
        "usage": {"completion_tokens": 0, "prompt_tokens": 0, "total_tokens": 0},
        "object": "chat.completion",
    }


def _construct_stream_resp_from_cache(return_message, saved_token):
    created = int(time.time())
    return [
        {
            "choices": [
                {"delta": {"role": "assistant"}, "finish_reason": None, "index": 0}
            ],
            "created": created,
            "object": "chat.completion.chunk",
        },
        {
            "choices": [
                {
                    "delta": {"content": return_message},
                    "finish_reason": None,
                    "index": 0,
                }
            ],
            "created": created,
            "object": "chat.completion.chunk",
        },
        {
            "gptcache": True,
            "choices": [{"delta": {}, "finish_reason": "stop", "index": 0}],
            "created": created,
            "object": "chat.completion.chunk",
            "saved_token": saved_token,
        },
    ]


def _construct_text_from_cache(return_text):
    return {
        "gptcache": True,
        "choices": [
            {
                "text": return_text,
                "finish_reason": "stop",
                "index": 0,
            }
        ],
        "created": int(time.time()),
        "usage": {"completion_tokens": 0, "prompt_tokens": 0, "total_tokens": 0},
        "object": "text_completion",
    }


def _construct_image_create_resp_from_cache(image_data, response_format, size):
    import_pillow()
    from PIL import Image as PILImage  # pylint: disable=C0415

    img_bytes = base64.b64decode((image_data))
    img_file = BytesIO(img_bytes)  # convert image to file-like object
    img = PILImage.open(img_file)
    new_size = tuple(int(a) for a in size.split("x"))
    if new_size != img.size:
        img = img.resize(new_size)
        buffered = BytesIO()
        img.save(buffered, format="JPEG")
    else:
        buffered = img_file

    if response_format == "url":
        target_url = os.path.abspath(str(int(time.time())) + ".jpeg")
        with open(target_url, "wb") as f:
            f.write(buffered.getvalue())
        image_data = target_url
    elif response_format == "b64_json":
        image_data = base64.b64encode(buffered.getvalue()).decode("ascii")
    else:
        raise AttributeError(
            f"Invalid response_format: {response_format} is not one of ['url', 'b64_json']"
        )

    return {
        "gptcache": True,
        "created": int(time.time()),
        "data": [{response_format: image_data}],
    }


def _construct_audio_text_from_cache(return_text):
    return {
        "gptcache": True,
        "text": return_text,
    }


def _num_tokens_from_messages(messages):
    """Returns the number of tokens used by a list of messages."""
    tokens_per_message = 3
    tokens_per_name = 1

    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += token_counter(value)
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens



import numpy as np

from gptcache.embedding.base import BaseEmbedding
from gptcache.utils import import_paddlenlp, import_paddle

import_paddle()
import_paddlenlp()


import paddle  # pylint: disable=C0413
from paddlenlp.transformers import AutoModel, AutoTokenizer  # pylint: disable=C0413

class PaddleNLP(BaseEmbedding):
    """Generate sentence embedding for given text using pretrained models from PaddleNLP transformers.

    :param model: model name, defaults to 'ernie-3.0-medium-zh'.
    :type model: str

    Example:
        .. code-block:: python

            from gptcache.embedding import PaddleNLP

            test_sentence = 'Hello, world.'
            encoder = PaddleNLP(model='ernie-3.0-medium-zh')
            embed = encoder.to_embeddings(test_sentence)
    """

    def __init__(self, model: str = "ernie-3.0-medium-zh"):
        self.model = AutoModel.from_pretrained(model)
        self.model.eval()

        self.tokenizer = AutoTokenizer.from_pretrained(model)
        if not self.tokenizer.pad_token:
            self.tokenizer.pad_token = "<pad>"
        self.__dimension = None

    def to_embeddings(self, data, **_):
        """Generate embedding given text input

        :param data: text in string.
        :type data: str

        :return: a text embedding in shape of (dim,).
        """
        if not isinstance(data, list):
            data = [data]
        inputs = self.tokenizer(
            data, padding=True, truncation=True, return_tensors="pd"
        )
        outs = self.model(**inputs)[0]
        emb = self.post_proc(outs, inputs).squeeze(0).detach().numpy()
        return np.array(emb).astype("float32")

    def post_proc(self, token_embeddings, inputs):
        attention_mask = paddle.ones(inputs["token_type_ids"].shape)
        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(token_embeddings.shape).astype("float32")
        )
        sentence_embs = paddle.sum(
            token_embeddings * input_mask_expanded, 1
        ) / paddle.clip(input_mask_expanded.sum(1), min=1e-9)
        return sentence_embs

    @property
    def dimension(self):
        """Embedding dimension.

        :return: embedding dimension
        """
        if not self.__dimension:
            self.__dimension = len(self.to_embeddings("foo"))
        return self.__dimension




import pickle
from abc import abstractmethod, ABCMeta
from typing import List, Any, Optional, Union

import cachetools
import numpy as np
import requests

from gptcache.manager.eviction import EvictionBase
from gptcache.manager.eviction.distributed_cache import NoOpEviction
from gptcache.manager.eviction_manager import EvictionManager
from gptcache.manager.object_data.base import ObjectBase
from gptcache.manager.scalar_data.base import (
    CacheStorage,
    CacheData,
    DataType,
    Answer,
    Question,
)
from gptcache.manager.vector_data.base import VectorBase, VectorData
from gptcache.utils.error import CacheError, ParamError
from gptcache.utils.log import gptcache_log


class DataManager(metaclass=ABCMeta):
    """DataManager manage the cache data, including save and search"""

    @abstractmethod
    def save(self, question, answer, embedding_data, **kwargs):
        pass

    @abstractmethod
    def import_data(
            self,
            questions: List[Any],
            answers: List[Any],
            embedding_datas: List[Any],
            session_ids: List[Optional[str]],
    ):
        pass

    @abstractmethod
    def get_scalar_data(self, res_data, **kwargs) -> CacheData:
        pass

    def hit_cache_callback(self, res_data, **kwargs):
        pass

    @abstractmethod
    def search(self, embedding_data, **kwargs):
        """search the data in the cache store accrodding to the embedding data

        :return: a list of search result, [[score, id], [score, id], ...]
        """
        pass

    def flush(self):
        pass

    @abstractmethod
    def add_session(self, res_data, session_id, pre_embedding_data):
        pass

    @abstractmethod
    def list_sessions(self, session_id, key):
        pass

    @abstractmethod
    def delete_session(self, session_id):
        pass

    def report_cache(
        self,
        user_question,
        cache_question,
        cache_question_id,
        cache_answer,
        similarity_value,
        cache_delta_time,
    ):
        pass

    @abstractmethod
    def close(self):
        pass


class MapDataManager(DataManager):
    """MapDataManager, store all data in a map data structure.

    :param data_path: the path to save the map data, defaults to 'data_map.txt'.
    :type data_path:  str
    :param max_size: the max size for the cache, defaults to 1000.
    :type max_size: int
    :param get_data_container: a Callable to get the data container, defaults to None.
    :type get_data_container:  Callable


    Example:
        .. code-block:: python

            from gptcache.manager import get_data_manager

            data_manager = get_data_manager("data_map.txt", 1000)
    """

    def __init__(self, data_path, max_size, get_data_container=None):
        if get_data_container is None:
            self.data = cachetools.LRUCache(max_size)
        else:
            self.data = get_data_container(max_size)
        self.data_path = data_path
        self.init()

    def init(self):
        try:
            with open(self.data_path, "rb") as f:
                self.data = pickle.load(f)
        except FileNotFoundError:
            return
        except PermissionError:
            raise CacheError(  # pylint: disable=W0707
                f"You don't have permission to access this file <{self.data_path}>."
            )

    def save(self, question, answer, embedding_data, **kwargs):
        if isinstance(question, Question):
            question = question.content
        session = kwargs.get("session", None)
        session_id = {session.name} if session else set()
        self.data[embedding_data] = (question, answer, embedding_data, session_id)

    def import_data(
        self,
        questions: List[Any],
        answers: List[Any],
        embedding_datas: List[Any],
        session_ids: List[Optional[str]],
    ):
        if (
            len(questions) != len(answers)
            or len(questions) != len(embedding_datas)
            or len(questions) != len(session_ids)
        ):
            raise ParamError("Make sure that all parameters have the same length")
        for i, embedding_data in enumerate(embedding_datas):
            self.data[embedding_data] = (
                questions[i],
                answers[i],
                embedding_datas[i],
                {session_ids[i]} if session_ids[i] else set(),
            )

    def get_scalar_data(self, res_data, **kwargs) -> CacheData:
        session = kwargs.get("session", None)
        if session:
            answer = (
                res_data[1].answer if isinstance(res_data[1], Answer) else res_data[1]
            )
            if not session.check_hit_func(
                session.name, list(res_data[3]), [res_data[0]], answer
            ):
                return None
        return CacheData(question=res_data[0], answers=res_data[1])

    def search(self, embedding_data, **kwargs):
        try:
            return [self.data[embedding_data]]
        except KeyError:
            return []

    def flush(self):
        try:
            with open(self.data_path, "wb") as f:
                pickle.dump(self.data, f)
        except PermissionError:
            gptcache_log.error(
                "You don't have permission to access this file %s.", self.data_path
            )

    def add_session(self, res_data, session_id, pre_embedding_data):
        res_data[3].add(session_id)

    def list_sessions(self, session_id=None, key=None):
        session_ids = set()
        for k in self.data:
            if session_id and session_id in self.data[k][3]:
                session_ids.add(k)
            elif len(self.data[k][3]) > 0:
                session_ids.update(self.data[k][3])
        return list(session_ids)

    def delete_session(self, session_id):
        keys = self.list_sessions(session_id=session_id)
        for k in keys:
            self.data[k][3].remove(session_id)
            if len(self.data[k][3]) == 0:
                del self.data[k]

    def close(self):
        self.flush()


def normalize(vec):
    magnitude = np.linalg.norm(vec)
    normalized_v = vec / magnitude
    return normalized_v


class SSDataManager(DataManager):
    """Generate SSDataManage to manager the data.

    :param s: CacheStorage to manager the scalar data, it can be generated with :meth:`gptcache.manager.CacheBase`.
    :type s: CacheStorage
    :param v: VectorBase to manager the vector data, it can be generated with :meth:`gptcache.manager.VectorBase`.
    :type v:  VectorBase
    :param o: ObjectBase to manager the object data, it can be generated with :meth:`gptcache.manager.ObjectBase`.
    :type o:  ObjectBase
    :param e: EvictionBase to manager the eviction data, it can be generated with :meth:`gptcache.manager.EvictionBase`.
    :type e:  EvictionBase
    """

    def __init__(
        self,
        s: CacheStorage,
        v: VectorBase,
        o: Optional[ObjectBase],
        e: Optional[EvictionBase],
        max_size,
        clean_size,
        policy="LRU"
    ):
        self.s = s
        self.v = v
        self.o = o
        self.eviction_manager = EvictionManager(self.s, self.v)
        if e is None:
            e = EvictionBase(name="memory",
                             maxsize=max_size,
                             clean_size=clean_size,
                             policy=policy,
                             on_evict=self._clear)
        self.eviction_base = e

        if not isinstance(self.eviction_base, NoOpEviction):
            # if eviction manager is no op redis, we don't need to put data into eviction base
            self.eviction_base.put(self.s.get_ids(deleted=False))

    def _clear(self, marked_keys):
        self.eviction_manager.soft_evict(marked_keys)
        if self.eviction_manager.check_evict():
            self.eviction_manager.delete()

    def save(self, question, answer, embedding_data, **kwargs):
        """Save the data and vectors to cache and vector storage.

        :param question: question data.
        :type question: str
        :param answer: answer data.
        :type answer: str, Answer or (Any, DataType)
        :param embedding_data: vector data.
        :type embedding_data: np.ndarray

        Example:
            .. code-block:: python

                import numpy as np
                from gptcache.manager import get_data_manager, CacheBase, VectorBase

                data_manager = get_data_manager(CacheBase('sqlite'), VectorBase('faiss', dimension=128))
                data_manager.save('hello', 'hi', np.random.random((128, )).astype('float32'))
        """
        session = kwargs.get("session", None)
        session_id = session.name if session else None
        self.import_data([question], [answer], [embedding_data], [session_id])

    def _process_answer_data(self, answers: Union[Answer, List[Answer]]):
        if isinstance(answers, Answer):
            answers = [answers]
        new_ans = []
        for ans in answers:
            if ans.answer_type != DataType.STR:
                new_ans.append(Answer(self.o.put(ans.answer), ans.answer_type))
            else:
                new_ans.append(ans)
        return new_ans

    def _process_question_data(self, question: Union[str, Question]):
        if isinstance(question, Question):
            if question.deps is None:
                return question

            for dep in question.deps:
                if dep.dep_type == DataType.IMAGE_URL:
                    dep.dep_type.data = self.o.put(requests.get(dep.data).content)
            return question

        return Question(question)

    def import_data(
        self,
        questions: List[Any],
        answers: List[Answer],
        embedding_datas: List[Any],
        session_ids: List[Optional[str]],
    ):
        if (
            len(questions) != len(answers)
            or len(questions) != len(embedding_datas)
            or len(questions) != len(session_ids)
        ):
            raise ParamError("Make sure that all parameters have the same length")
        cache_datas = []
        embedding_datas = [
            normalize(embedding_data) for embedding_data in embedding_datas
        ]
        for i, embedding_data in enumerate(embedding_datas):
            if self.o is not None and not isinstance(answers[i], str):
                ans = self._process_answer_data(answers[i])
            else:
                ans = answers[i]

            cache_datas.append(
                CacheData(
                    question=self._process_question_data(questions[i]),
                    answers=ans,
                    embedding_data=embedding_data.astype("float32"),
                    session_id=session_ids[i],
                )
            )
        ids = self.s.batch_insert(cache_datas)
        self.v.mul_add(
            [
                VectorData(id=ids[i], data=embedding_data)
                for i, embedding_data in enumerate(embedding_datas)
            ]
        )
        self.eviction_base.put(ids)

    def get_scalar_data(self, res_data, **kwargs) -> Optional[CacheData]:
        session = kwargs.get("session", None)
        cache_data = self.s.get_data_by_id(res_data[1])
        if cache_data is None:
            return None

        if session:
            cache_answer = (
                cache_data.answers[0].answer
                if isinstance(cache_data.answers[0], Answer)
                else cache_data.answers[0]
            )
            res_list = self.list_sessions(key=res_data[1])
            cache_session_ids, cache_questions = [r.session_id for r in res_list], [
                r.session_question for r in res_list
            ]
            if not session.check_hit_func(
                session.name, cache_session_ids, cache_questions, cache_answer
            ):
                return None

        for ans in cache_data.answers:
            if ans.answer_type != DataType.STR:
                ans.answer = self.o.get(ans.answer)
        return cache_data

    def hit_cache_callback(self, res_data, **kwargs):
        self.eviction_base.get(res_data[1])

    def search(self, embedding_data, **kwargs):
        embedding_data = normalize(embedding_data)
        top_k = kwargs.get("top_k", -1)
        return self.v.search(data=embedding_data, top_k=top_k)

    def flush(self):
        self.s.flush()
        self.v.flush()

    def add_session(self, res_data, session_id, pre_embedding_data):
        self.s.add_session(res_data[1], session_id, pre_embedding_data)

    def list_sessions(self, session_id=None, key=None):
        res = self.s.list_sessions(session_id, key)
        if key:
            return res
        if session_id:
            return list(r.id for r in res)
        return list(set(r.session_id for r in res))

    def delete_session(self, session_id):
        keys = self.list_sessions(session_id=session_id)
        self.s.delete_session(keys)

    def report_cache(
        self,
        user_question,
        cache_question,
        cache_question_id,
        cache_answer,
        similarity_value,
        cache_delta_time,
    ):
        self.s.report_cache(
            user_question,
            cache_question,
            cache_question_id,
            cache_answer,
            similarity_value,
            cache_delta_time,
        )

    def close(self):
        self.s.close()
        self.v.close()



__all__ = ["ObjectBase"]

from gptcache.utils.lazy_import import LazyImport

object_manager = LazyImport(
    "object_manager", globals(), "gptcache.manager.object_data.manager"
)


def ObjectBase(name: str, **kwargs):
    """Generate specific ObjectStorage with the configuration. For example, setting for
       `ObjectBase` (with `name`) to manage LocalObjectStorage, S3 object storage.
    """
    return object_manager.ObjectBase.get(name, **kwargs)



from functools import reduce
from random import randint, SystemRandom
from datetime import datetime
from typing import List, Optional, Dict
from decimal import Decimal
from gptcache.manager.scalar_data.base import CacheStorage, CacheData, Question, QuestionDep, Answer
from gptcache.utils import import_boto3
from gptcache.utils.log import gptcache_log
from concurrent.futures import ThreadPoolExecutor, wait
import numpy as np
import math

import_boto3()
from boto3.session import Session as AwsSession
from boto3.dynamodb.conditions import Key as DynamoKey, Attr as DynamoAttr
from boto3.dynamodb.types import Binary as DynamoBinary

class DynamoStorage(CacheStorage):
    """
    DynamoDB storage using AWS's boto3 library.

    :param aws_access_key_id: AWS access key ID. If not specified, boto3 will use the default resolution behavior.
    :type host: str

    :param aws_secret_access_key: AWS secret access key. If not specified, boto3 will use the default resolution behavior.
    :type aws_secret_access_key str

    :param aws_region_name: AWS region name. If not specified, boto3 will use the default resolution behavior.
    :type aws_region_name: str

    :param aws_profile_name: AWS profile name. If not specified, boto3 will use the default resolution behavior.
    :type aws_profile_name: str

    :param aws_endpoint_url: AWS endpoint URL. This is normally handled automatically but is exposed to allow overriding for testing purposes
                            (using something like LocalStack or dynamoDB-local for example).
    :type aws_endpoint_url: str
    """

    max_cardinality_suffix = 10

    def __init__(
        self,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        aws_region_name: str = None,
        aws_profile_name: str = None,
        aws_endpoint_url: str = None,
    ):
        self._aws_session = AwsSession(
            aws_access_key_id = aws_access_key_id,
            aws_secret_access_key = aws_secret_access_key,
            region_name = aws_region_name,
            profile_name = aws_profile_name,
        )
        self._dynamo = self._aws_session.resource(
            "dynamodb",
            endpoint_url=aws_endpoint_url,
        )
        self.create()

    def create(self):
        # NOTE: We use PAY_PER_REQUEST billing mode as per AWS's recommendations for unpredicatble workloads since it's
        #       being used as a cache and so we won't know how many reads/writes we'll need to perform ahead of time.
        recommended_billing_mode = "PAY_PER_REQUEST"

        if not self._does_table_already_exist_and_is_active("gptcache_questions"):
            self._dynamo.create_table(
                TableName = "gptcache_questions",
                KeySchema = [
                    {"AttributeName": "pk", "KeyType": "HASH"},
                    {"AttributeName": "id", "KeyType": "RANGE"},
                ],
                AttributeDefinitions = [
                    {"AttributeName": "pk", "AttributeType": "S"},
                    {"AttributeName": "id", "AttributeType": "S"},

                    # You might be wondering why the 'deleted' attribute is a string value, not a BOOL.
                    # This is because we need to efficiently be able to query items that are soft-deleted. This is normally
                    # done with a GSI (GlobalSecondaryIndex). However, GSIs don't support the BOOL type due to low cardinality.
                    # So we use a string instead.
                    {"AttributeName": "deleted", "AttributeType": "S"},
                ],
                GlobalSecondaryIndexes = [
                    {
                        "IndexName": "gsi_items_by_type",
                        "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
                        "Projection": {"ProjectionType": "ALL"},
                    },
                    {
                        "IndexName": "gsi_questions_by_deletion_status",
                        "KeySchema": [{"AttributeName": "deleted", "KeyType": "HASH"}],
                        "Projection": {"ProjectionType": "ALL"},
                    },
                ],
                BillingMode = recommended_billing_mode,
            )

        if not self._does_table_already_exist_and_is_active("gptcache_reports"):
            self._dynamo.create_table(
                TableName = "gptcache_reports",
                KeySchema = [
                    {"AttributeName": "id", "KeyType": "HASH"},
                ],
                AttributeDefinitions = [
                    {"AttributeName": "id", "AttributeType": "S"},
                ],
                BillingMode = recommended_billing_mode,
            )

    def batch_insert(self, all_data: List[CacheData]) -> List[str]:
        """
        Inserts a list of CacheData objects into the DynamoDB table and returns the ids of the inserted rows
        """
        table = self._dynamo.Table("gptcache_questions")
        table.wait_until_exists()
        ids = []

        with table.batch_writer() as batch:
            for item in all_data:
                item_deps = item.question.deps if isinstance(item.question, Question) and item.question.deps is not None else []
                new_id_without_prefix = self._generate_id()
                new_id = f"questions#{new_id_without_prefix}"
                ids.append(new_id_without_prefix)
                creation_time = item.create_on if item.create_on is not None else datetime.utcnow()

                batch.put_item(
                    Item = self._strip_all_props_with_none_values({
                        "pk": new_id,
                        "id": new_id,
                        "question": self._question_text(item.question),
                        "answers": [self._serialize_answer(answer) for answer in item.answers],
                        "deps": [self._serialize_question_deps(dep) for dep in item_deps],
                        #"session_id": item.session_id if item.session_id is not None else None,
                        "create_on": creation_time.isoformat(timespec="microseconds"),
                        "last_access": item.last_access.isoformat(timespec="microseconds") if item.last_access is not None else None,
                        "embedding_data": DynamoBinary(item.embedding_data.tobytes()) if item.embedding_data is not None else None,
                        "deleted": self._serialize_deleted_value(False),
                    })
                )

                if item.session_id:
                    batch.put_item(
                        Item = self._strip_all_props_with_none_values({
                            "pk": new_id,
                            "id": f"sessions#{item.session_id}",
                            "question": self._question_text(item.question),
                        })
                    )
        return ids

    def get_data_by_id(self, key: str) -> Optional[CacheData]:
        table = self._dynamo.Table("gptcache_questions")
        table.wait_until_exists()

        key_with_prefix = f"questions#{key}"

        # get all sessions that contain this question
        sessions_containing_question = table.query(
            KeyConditionExpression = DynamoKey("pk").eq(key_with_prefix) & DynamoKey("id").begins_with("sessions#"),
        )["Items"]

        # then get the question info itself
        response = table.get_item(
            Key={"pk": key_with_prefix, "id": key_with_prefix},
        )

        if "Item" not in response or self._deserialize_deleted_value(response["Item"]["deleted"]):
            return None

        # after it's accessed, we need to update that particular item's last_access timestamp
        table.update_item(
            Key={"pk": key_with_prefix, "id": key_with_prefix},
            UpdateExpression="SET last_access = :last_access",
            ExpressionAttributeValues={
                ":last_access": datetime.utcnow().isoformat(timespec="microseconds"),
            },
        )

        cache_data = self._response_item_to_cache_data(response["Item"])
        cache_data.session_id = [session["id"].split("#")[1] for session in sessions_containing_question]
        return cache_data

    def mark_deleted(self, keys: str):
        table = self._dynamo.Table("gptcache_questions")
        table.wait_until_exists()

        with table.batch_writer() as batch:
            for key in keys:
                batch.put_item(
                    Item={
                        "pk": f"questions#{key}",
                        "id": f"questions#{key}",
                        "deleted": self._serialize_deleted_value(True),
                    }
                )

    def clear_deleted_data(self):
        # Since the 'deleted' attribute is a string, we need to query for all items that have a 'deleted' value of
        # 'True_*'. Then we can call delete on them.
        table = self._dynamo.Table("gptcache_questions")
        table.wait_until_exists()

        # Build out all possible values for the 'deleted' attribute and query it in 1 go. It cannot be done as one
        # network call unfortunately due to limitations of DynamoDB. The DB only allows running queries on the GSI
        # with a SINGLE, concrete value for the partition key.
        #
        # To address this, we're gonna spin up a thread pool and query concurrently
        with ThreadPoolExecutor(max_workers = 10) as executor:
            futures = []
            for i in range(1, self.max_cardinality_suffix + 1):
                futures.append(executor.submit(
                    table.query,
                    IndexName = "gsi_questions_by_deletion_status",
                    KeyConditionExpression = DynamoKey("deleted").eq(f"True_{i}")
                ))

            completed_responses, incomplete_responses = wait(futures, timeout = 10)

            if len(incomplete_responses) > 0:
                gptcache_log.error(
                    """
                    Unable to complete deletion of all soft-deleted items due to request timeout.
                    Some items may remain in the cache. %s",
                    """,
                    incomplete_responses,
                )

        soft_deleted_entries_per_partition = [response.result()["Items"] for response in completed_responses]
        soft_deleted_entries = reduce(lambda x, y: x + y, soft_deleted_entries_per_partition)

        keys = [entry["id"] for entry in soft_deleted_entries]

        with table.batch_writer() as batch:
            for key in keys:
                batch.delete_item(
                    Key={"pk": key, "id": key},
                )

    def get_ids(self, deleted: bool = True) -> List[str]:
        table = self._dynamo.Table("gptcache_questions")
        table.wait_until_exists()

        def run_scan_operation(last_evaluated_key):
            filter_expression = (
                DynamoAttr("id").begins_with("questions#") & DynamoAttr("deleted").begins_with(f"{deleted}_")
            )

            if last_evaluated_key is None:
                return table.scan(
                    Select = "SPECIFIC_ATTRIBUTES",
                    ProjectionExpression = "pk",
                    FilterExpression = filter_expression,
                )
            return table.scan(
                Select = "SPECIFIC_ATTRIBUTES",
                ProjectionExpression = "pk",
                FilterExpression = filter_expression,
                ExclusiveStartKey = last_evaluated_key,
            )

        all_responses = self._fetch_all_pages(run_scan_operation)
        all_items = reduce(lambda x, y: x + y, [response["Items"] for response in all_responses])
        all_ids = [int(item["pk"].replace("questions#", "")) for item in all_items]
        return all_ids

    def count(self, state: int = 0, is_all: bool = False) -> int:
        table = self._dynamo.Table("gptcache_questions")
        table.wait_until_exists()
        key_condition_expression = DynamoAttr("id").begins_with("questions#")

        if not is_all and state == 0:
            key_condition_expression &= DynamoAttr("deleted").begins_with("False_")
        elif not is_all and state != 0:
            key_condition_expression &= DynamoAttr("deleted").begins_with("True_")

        # TODO: find out if specifying a "COUNT" select type results in a single page or not. For now, assume the worst
        #       and attempt pagination anyway.
        def run_scan_operation(last_evaluated_key):
            if last_evaluated_key is None:
                return table.scan(
                    IndexName = "gsi_questions_by_deletion_status",
                    FilterExpression = key_condition_expression,
                    Select = "COUNT",
                )

            return table.scan(
                IndexName = "gsi_questions_by_deletion_status",
                FilterExpression = key_condition_expression,
                ExclusiveStartKey = last_evaluated_key,
                Select = "COUNT",
            )

        all_responses = self._fetch_all_pages(run_scan_operation)
        return sum([response["Count"] for response in all_responses])

    def add_session(self, question_id: str, session_id: str, session_question: str):
        table = self._dynamo.Table("gptcache_questions")
        table.wait_until_exists()

        table.put_item(
            Item = {
                "pk": f"questions#{question_id}",
                "id": f"sessions#{session_id}",
                "question": session_question,
            },
        )

    def list_sessions(self, session_id = None, key = None) -> List[str]:
        table = self._dynamo.Table("gptcache_questions")
        table.wait_until_exists()

        if session_id and key:
            response = table.query(
                IndexName = "gsi_items_by_type",
                ProjectionExpression = "id",
                KeyConditionExpression = DynamoKey("id").eq(f"sessions#{session_id}"),
                FilterExpression = DynamoAttr("pk").eq(f"questions#{key}"),
            )
        elif session_id:
            response = table.query(
                IndexName = "gsi_items_by_type",
                ProjectionExpression = "id",
                KeyConditionExpression = DynamoKey("id").eq(f"sessions#{session_id}"),
            )
        elif key:
            response = table.query(
                ProjectionExpression = "id",
                KeyConditionExpression = DynamoKey("pk").eq(f"questions#{key}") & DynamoKey("id").begins_with("sessions#"),
            )
        else:
            def run_scan_operation(last_evaluated_key):
                filter_expression = DynamoAttr("id").begins_with("sessions#")

                if last_evaluated_key is None:
                    return table.scan(
                        ProjectionExpression = "id",
                        FilterExpression = filter_expression,
                    )
                return table.scan(
                    ProjectionExpression = "id",
                    FilterExpression = filter_expression,
                    ExclusiveStartKey = last_evaluated_key,
                )

            response = self._fetch_all_pages(run_scan_operation)

            # since these are paginated results, merge all items in each response to a new dict
            response = reduce(lambda x, y: { "Items": x["Items"] + y["Items"] }, response)

        # since sessions can be the shared across multiple items, we need to dedupe the results
        return list({ item["id"].replace("questions#", "") for item in response["Items"] })

    def delete_session(self, keys: List[str]):
        table = self._dynamo.Table("gptcache_questions")
        table.wait_until_exists()

        # first find all items with that session_id
        # unfortunately, there is no "batch get" operation on a GSI. So we need to spin up a thread pool and query
        # concurrently
        with ThreadPoolExecutor(max_workers = 10) as executor:
            futures = []
            for key in keys:
                futures.append(executor.submit(
                    table.query,
                    IndexName = "gsi_items_by_type",
                    Select = "SPECIFIC_ATTRIBUTES",
                    ProjectionExpression = "pk, id",
                    KeyConditionExpression = DynamoKey("id").eq(f"sessions#{key}")
                ))

            completed_responses, incomplete_responses = wait(futures, timeout = 10)

            if len(incomplete_responses) > 0:
                gptcache_log.error(
                    "Unable to query all questions in session due to request timeout. %s",
                    incomplete_responses,
                )

        items_per_session = [response.result()["Items"] for response in completed_responses]
        all_relevant_items = reduce(lambda x, y: x + y, items_per_session)

        # now we need to delete all items with those keys to clear out all session data.
        with table.batch_writer() as batch:
            for item in all_relevant_items:
                batch.delete_item(
                    Key={"pk": item["pk"], "id": item["id"]},
                )

    def report_cache(
        self,
        user_question,
        cache_question,
        cache_question_id,
        cache_answer,
        similarity_value,
        cache_delta_time,
    ):
        table = self._dynamo.Table("gptcache_reports")
        table.wait_until_exists()

        table.put_item(
            Item={
                "id": str(self._generate_id()),
                "user_question": user_question,
                "cache_question": cache_question,
                "cache_question_id": cache_question_id,
                "cache_answer": cache_answer,
                "similarity": (
                    # DynamoDB doesn't support floats; only decimals
                    Decimal(str(similarity_value))
                    if isinstance(similarity_value, float)
                    else similarity_value
                ),
                "cache_delta_time": Decimal(str(cache_delta_time)),
            }
        )

    def close(self):
        pass

    def _does_table_already_exist_and_is_active(self, table_name: str) -> bool:
        try:
            self._dynamo.Table(table_name).table_status in ["ACTIVE", "UPDATING"]
        except self._dynamo.meta.client.exceptions.ResourceNotFoundException:
            return False
        return True

    def _wait_until_table_exists(self, table_name: str):
        """
        When tables are created in DynamoDB, they are not immediately available as table creation is asynchronous.
        This function will block until the table is available in order to ensure we don't attempt to perform any
        operations on a table that doesn't exist yet.
        """
        self._dynamo.Table(table_name).wait_until_exists()


    def _response_item_to_cache_data(self, question_resp: Dict) -> Optional[CacheData]:
        deps = question_resp["deps"] if "deps" in question_resp else []
        return CacheData(
            question = Question(
                content = question_resp["question"],
                deps = [self._deserialize_question_dep(dep) for dep in deps]
            ),
            answers = [self._deserialize_answer(answer) for answer in question_resp["answers"]],
            embedding_data = np.frombuffer(question_resp["embedding_data"].value, dtype=np.float32),
            create_on = datetime.fromisoformat(question_resp["create_on"]),
            last_access = datetime.fromisoformat(question_resp["last_access"]) if "last_access" in question_resp else None,
        )

    def _fetch_all_pages(self, scan_fn) -> List[Dict]:
        """
        We often have to resort to scan operations in Dynamo which could result in a lot of data being returned. To ensure,
        we get all the data, we need to paginate through the results. This function handles that for us.

        :param scan_fn: The function to call to perform the scan operation. It must take in a last_evaluated_key to pass
                        along to Dynamo and return a Dynamo response object

        see: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/table/scan.html
        """
        all_responses = []
        last_evaluated_key = None

        while True:
            response = scan_fn(last_evaluated_key)
            all_responses.append(response)

            if "LastEvaluatedKey" in response and len(response["LastEvaluatedKey"].keys()) > 0:
                last_evaluated_key = response["LastEvaluatedKey"]
            else:
                break

        return all_responses

    def _question_text(self, question) -> str:
        return question if isinstance(question, str) else question.content

    def _generate_id(self):
        """
        Generates a unique ID for a row. Unfortunately, we cannot use something like uuid4() as scalar storage
        implmentations need to return an integer based id.

        It seems the underlying adapter that uses this class converts the ids to C longs which are 64-bit signed
        so we'll use that as our upper bound.

        We use systemrandom for better randomness.
        """
        return SystemRandom().randint(1, math.pow(2, 63) - 1)

    def _strip_all_props_with_none_values(self, obj: Dict) -> Dict:
        sanitized = {}
        for k, v in obj.items():
            if v is not None:
                sanitized[k] = v
        return sanitized

    def _serialize_deleted_value(self, value: bool, suffix_value = randint(1, max_cardinality_suffix)) -> str:
        """
        We need to be able to query and filter on the 'deleted' attribute. However, in order for us to use this value
        as an indexable value, it cannot be a BOOL (DynamoDB doesn't support it due to low cardinaility).

        To increase the cardinality, we first convert the boolean to a string and append a random integer.

        see: https://aws.amazon.com/blogs/database/how-to-design-amazon-dynamodb-global-secondary-indexes/
        """
        return f"{value}_{suffix_value}"

    def _serialize_answer(self, answer: Answer) -> Dict:
        return { "answer": answer.answer, "answer_type": answer.answer_type.value }

    def _serialize_question_deps(self, dep: QuestionDep) -> Dict:
        return { "name": dep.name, "data": dep.data, "dep_type": dep.dep_type.value }

    def _deserialize_deleted_value(self, value: str) -> bool:
        return value.startswith("True_")

    def _deserialize_question_dep(self, dep: Dict) -> QuestionDep:
        return QuestionDep(name = dep["name"], data = dep["data"], dep_type = int(dep["dep_type"]))

    def _deserialize_answer(self, answer: Dict) -> Answer:
        return Answer(answer = answer["answer"], answer_type = int(answer["answer_type"]))



import os

import numpy as np

from gptcache.embedding.base import BaseEmbedding
from gptcache.utils import import_openai

import_openai()

import openai  # pylint: disable=C0413

class OpenAI(BaseEmbedding):
    """Generate text embedding for given text using OpenAI.

    :param model: model name, defaults to 'text-embedding-ada-002'.
    :type model: str
    :param api_key: OpenAI API Key. When the parameter is not specified, it will load the key by default if it is available.
    :type api_key: str

    Example:
        .. code-block:: python

            from gptcache.embedding import OpenAI

            test_sentence = 'Hello, world.'
            encoder = OpenAI(api_key='your_openai_key')
            embed = encoder.to_embeddings(test_sentence)
    """

    def __init__(self, model: str = "text-embedding-ada-002", api_key: str = None, api_base: str = None):
        if not api_key:
            if openai.api_key:
                api_key = openai.api_key
            else:
                api_key = os.getenv("OPENAI_API_KEY")
        if not api_base:
            if openai.api_base:
                api_base = openai.api_base
            else:
                api_base = os.getenv("OPENAI_API_BASE")
        openai.api_key = api_key
        self.api_base = api_base  # don't override all of openai as we may just want to override for say embeddings
        self.model = model
        if model in self.dim_dict():
            self.__dimension = self.dim_dict()[model]
        else:
            self.__dimension = None

    def to_embeddings(self, data, **_):
        """Generate embedding given text input

        :param data: text in string.
        :type data: str

        :return: a text embedding in shape of (dim,).
        """
        sentence_embeddings = openai.Embedding.create(model=self.model, input=data, api_base=self.api_base)
        return np.array(sentence_embeddings["data"][0]["embedding"]).astype("float32")

    @property
    def dimension(self):
        """Embedding dimension.

        :return: embedding dimension
        """
        if not self.__dimension:
            foo_emb = self.to_embeddings("foo")
            self.__dimension = len(foo_emb)
        return self.__dimension

    @staticmethod
    def dim_dict():
        return {"text-embedding-ada-002": 1536}



__all__ = ["CacheBase"]

from gptcache.utils.lazy_import import LazyImport

scalar_manager = LazyImport(
    "scalar_manager", globals(), "gptcache.manager.scalar_data.manager"
)


def CacheBase(name: str, **kwargs):
    """Generate specific CacheStorage with the configuration. For example, setting for
       `SQLDataBase` (with `name`, `sql_url` and `table_name` params) to manage SQLite, PostgreSQL, MySQL, MariaDB, SQL Server and Oracle.
    """
    return scalar_manager.CacheBase.get(name, **kwargs)



from typing import Any, Dict

from gptcache.processor import ContextProcess


class ConcatContextProcess(ContextProcess):
    """A concat context processor simply concat the context.
    Generally used with rwkv embedding, because rwkv can input almost infinitely long

    Example:
        .. code-block:: python

            from gptcache.manager import manager_factory
            from gptcache.processor.context.concat_context import ConcatContextProcess

            context_process = ConcatContextProcess()
            rwkv_embedding = Rwkv()
            data_manager = manager_factory(
                "sqlite,faiss",
                vector_params={"dimension": rwkv_embedding.dimension},
            )
            cache.init(
                pre_embedding_func=context_process.pre_process,
                embedding_func=rwkv_embedding.to_embeddings,
                data_manager=data_manager,
            )
    """

    content: str = ""

    def __init__(
            self
    ):
        self.content = ""
        self.concat_content = ""

    def format_all_content(self, data: Dict[str, Any], **params: Dict[str, Any]):
        for query in data["messages"]:
            self.content += f"{query['role']}: {query['content']} \n"
            self.concat_content += query["content"]

    def process_all_content(self) -> (Any, Any):
        return self.content, self.concat_content



class EvictionManager:
    """
    EvictionManager to manager the eviction policy.

    :param scalar_storage: CacheStorage to manager the scalar data.
    :type scalar_storage: :class:`CacheStorage`
    :param vector_base: VectorBase to manager the vector data.
    :type vector_base:  :class:`VectorBase`
    """

    MAX_MARK_COUNT = 5000
    MAX_MARK_RATE = 0.1
    BATCH_SIZE = 100000
    REBUILD_CONDITION = 5

    def __init__(self, scalar_storage, vector_base):
        self._scalar_storage = scalar_storage
        self._vector_base = vector_base
        self.delete_count = 0

    def check_evict(self):
        mark_count = self._scalar_storage.count(state=-1)
        all_count = self._scalar_storage.count(is_all=True)
        if (
            mark_count > self.MAX_MARK_COUNT
            or mark_count / all_count > self.MAX_MARK_RATE
        ):
            return True
        return False

    def delete(self):
        mark_ids = self._scalar_storage.get_ids(deleted=True)
        self._scalar_storage.clear_deleted_data()
        self._vector_base.delete(mark_ids)
        self.delete_count += 1
        if self.delete_count >= self.REBUILD_CONDITION:
            self.rebuild()

    def rebuild(self):
        self._scalar_storage.clear_deleted_data()
        ids = self._scalar_storage.get_ids(deleted=False)
        self._vector_base.rebuild(ids)
        self.delete_count = 0

    def soft_evict(self, marked_keys):
        self._scalar_storage.mark_deleted(marked_keys)



# pylint: disable=wrong-import-position
from abc import ABC
from typing import List

from gptcache.manager.eviction.distributed_cache import DistributedEviction
from gptcache.utils import import_redis


import_redis()
import redis
from redis_om import get_redis_connection


class RedisCacheEviction(DistributedEviction, ABC):
    """eviction: Distributed Cache Eviction Strategy using Redis.

    :param host: the host of redis
    :type host: str
    :param port: the port of redis
    :type port: int
    :param policy: eviction strategy policy of redis such as allkeys-lru, volatile-lru, allkeys-random, volatile-random, etc.
    refer https://redis.io/docs/reference/eviction/ for more information.
    :type policy: str
    :param maxsize: the maxsize of cache data
    :type maxsize: int
    :param on_evict: the function for cleaning the data in the store
    :type  on_evict: Callable[[List[Any]], None]
    :param maxmemory: the maxmemory of redis
    :type maxmemory: str
    :param global_key_prefix: the global key prefix
    :type global_key_prefix: str
    :param ttl: the ttl of the cache data
    :type ttl: int
    :param maxmemory_samples: Number of keys to sample when evicting keys
    :type maxmemory_samples: int
    :param kwargs: the kwargs
    :type kwargs: Any
    """

    def __init__(self,
                 host="localhost",
                 port=6379,
                 maxmemory: str = None,
                 policy: str = None,
                 global_key_prefix="gptcache",
                 ttl: int = None,
                 maxmemory_samples: int = None,
                 **kwargs):
        self._redis = get_redis_connection(host=host, port=port, **kwargs)
        if maxmemory:
            self._redis.config_set("maxmemory", maxmemory)
        if maxmemory_samples:
            self._redis.config_set("maxmemory-samples", maxmemory_samples)
        if policy:
            self._redis.config_set("maxmemory-policy", policy)
            self._policy = policy.lower()

        self._global_key_prefix = global_key_prefix
        self._ttl = ttl

    def _create_key(self, key: str) -> str:
        return f"{self._global_key_prefix}:evict:{key}"

    def put(self, objs: List[str], expire=False):
        ttl = self._ttl if expire else None
        for key in objs:
            self._redis.set(self._create_key(key), "True", ex=ttl)

    def get(self, obj: str):

        try:
            value = self._redis.get(self._create_key(obj))
            # update key expire time when accessed
            if self._ttl:
                self._redis.expire(self._create_key(obj), self._ttl)
            return value
        except redis.RedisError:
            print(f"Error getting key {obj} from cache")
            return None

    @property
    def policy(self) -> str:
        return self._policy



from abc import ABCMeta, abstractmethod


class BaseEmbedding(metaclass=ABCMeta):
    """
    Base Embedding interface.
    """

    @abstractmethod
    def to_embeddings(self, data, **kwargs):
        pass

    @property
    @abstractmethod
    def dimension(self) -> int:
        return 0



from typing import List, Optional, Tuple, Union
import numpy as np

from gptcache.utils import import_weaviate
from gptcache.utils.log import gptcache_log
from gptcache.manager.vector_data.base import VectorBase, VectorData

import_weaviate()

from weaviate import Client
from weaviate.auth import AuthCredentials
from weaviate.config import Config
from weaviate.embedded import EmbeddedOptions
from weaviate.types import NUMBERS


class Weaviate(VectorBase):
    """
    vector store: Weaviate
    """

    TIMEOUT_TYPE = Union[Tuple[NUMBERS, NUMBERS], NUMBERS]

    def __init__(
        self,
        url: Optional[str] = None,
        auth_client_secret: Optional[AuthCredentials] = None,
        timeout_config: TIMEOUT_TYPE = (10, 60),
        proxies: Union[dict, str, None] = None,
        trust_env: bool = False,
        additional_headers: Optional[dict] = None,
        startup_period: Optional[int] = 5,
        embedded_options: Optional[EmbeddedOptions] = None,
        additional_config: Optional[Config] = None,
        class_name: str = "GPTCache",
        class_schema: dict = None,
        top_k: Optional[int] = 1,
    ) -> None:

        if url is None and embedded_options is None:
            embedded_options = EmbeddedOptions()

        self.client = Client(
            url=url,
            auth_client_secret=auth_client_secret,
            timeout_config=timeout_config,
            proxies=proxies,
            trust_env=trust_env,
            additional_headers=additional_headers,
            startup_period=startup_period,
            embedded_options=embedded_options,
            additional_config=additional_config,
        )

        if class_schema:
            self.class_schema = class_schema
            self.class_name = class_schema.get("class")
        else:
            self.class_name = class_name
            self.class_schema = self._get_default_class_schema()

        self._create_class()
        self.top_k = top_k

    def _create_class(self):
        if self.client.schema.exists(self.class_name):
            gptcache_log.warning(
                "The %s collection already exists, and it will be used directly.",
                self.class_name,
            )
        else:
            self.client.schema.create_class(self.class_schema)
        return self.class_name

    def _get_default_class_schema(self) -> dict:
        return {
            "class": self.class_name,
            "description": "LLM response cache",
            "properties": [
                {
                    "name": "data_id",
                    "dataType": ["int"],
                    "description": "The data-id generated by GPTCache for vectors.",
                }
            ],
            "vectorIndexConfig": {"distance": "cosine"},
        }

    def mul_add(self, datas: List[VectorData]):
        with self.client.batch(batch_size=100, dynamic=True) as batch:
            for data in datas:
                properties = {
                    "data_id": data.id,
                }

                batch.add_data_object(
                    data_object=properties, class_name=self.class_name, vector=data.data
                )

    def search(self, data: np.ndarray, top_k: int = -1):
        if top_k == -1:
            top_k = self.top_k

        result = (
            self.client.query.get(class_name=self.class_name, properties=["data_id"])
            .with_near_vector(content={"vector": data})
            .with_additional(["distance"])
            .with_limit(top_k)
            .do()
        )

        return list(
            map(
                lambda x: (x["_additional"]["distance"], x["data_id"]),
                result["data"]["Get"][self.class_name],
            )
        )

    def _get_uuids(self, data_ids):
        uuid_list = []

        for data_id in data_ids:
            res = (
                self.client.query.get(
                    class_name=self.class_name, properties=["data_id"]
                )
                .with_where(
                    {"path": ["data_id"], "operator": "Equal", "valueInt": data_id}
                )
                .with_additional(["id"])
                .do()
            )

            uuid_list.append(
                res["data"]["Get"][self.class_name][0]["_additional"]["id"]
            )

        return uuid_list

    def delete(self, ids):
        uuids = self._get_uuids(ids)

        for uuid in uuids:
            self.client.data_object.delete(class_name=self.class_name, uuid=uuid)

    def rebuild(self, ids=None):
        return

    def flush(self):
        self.client.batch.flush()

    def close(self):
        self.flush()

    def get_embeddings(self, data_id: int):
        results = (
            self.client.query.get(class_name=self.class_name, properties=["data_id"])
            .with_where(
                {
                    "path": ["data_id"],
                    "operator": "Equal",
                    "valueInt": data_id,
                }
            )
            .with_additional(["vector"])
            .with_limit(1)
            .do()
        )

        results = results["data"]["Get"][self.class_name]

        if len(results) < 1:
            return None

        vec_emb = np.asarray(results[0]["_additional"]["vector"], dtype="float32")
        return vec_emb

    def update_embeddings(self, data_id: int, emb: np.ndarray):
        self.delete([data_id])

        properties = {
            "data_id": data_id,
        }

        self.client.data_object.create(
            data_object=properties, class_name=self.class_name, vector=emb
        )



import atexit
import os
from typing import Optional, List, Any

from gptcache.config import Config
from gptcache.embedding.string import to_embeddings as string_embedding
from gptcache.manager import get_data_manager
from gptcache.manager.data_manager import DataManager
from gptcache.processor.post import temperature_softmax
from gptcache.processor.pre import last_content
from gptcache.report import Report
from gptcache.similarity_evaluation import ExactMatchEvaluation
from gptcache.similarity_evaluation import SimilarityEvaluation
from gptcache.utils import import_openai
from gptcache.utils.cache_func import cache_all
from gptcache.utils.log import gptcache_log


class Cache:
    """GPTCache core object.


    Example:
        .. code-block:: python

            from gptcache import cache
            from gptcache.adapter import openai

            cache.init()
            cache.set_openai_key()
    """

    # it should be called when start the cache system
    def __init__(self):
        self.has_init = False
        self.cache_enable_func = None
        self.pre_embedding_func = None
        self.embedding_func = None
        self.data_manager: Optional[DataManager] = None
        self.similarity_evaluation: Optional[SimilarityEvaluation] = None
        self.post_process_messages_func = None
        self.config = Config()
        self.report = Report()
        self.next_cache = None

    def init(
        self,
        cache_enable_func=cache_all,
        pre_embedding_func=last_content,
        pre_func=None,
        embedding_func=string_embedding,
        data_manager: DataManager = get_data_manager(),
        similarity_evaluation=ExactMatchEvaluation(),
        post_process_messages_func=temperature_softmax,
        post_func=None,
        config=Config(),
        next_cache=None,
    ):
        """Pass parameters to initialize GPTCache.

        :param cache_enable_func: a function to enable cache, defaults to ``cache_all``
        :param pre_embedding_func: a function to preprocess embedding, defaults to ``last_content``
        :param pre_func: a function to preprocess embedding, same as ``pre_embedding_func``
        :param embedding_func: a function to extract embeddings from requests for similarity search, defaults to ``string_embedding``
        :param data_manager: a ``DataManager`` module, defaults to ``get_data_manager()``
        :param similarity_evaluation: a module to calculate embedding similarity, defaults to ``ExactMatchEvaluation()``
        :param post_process_messages_func: a function to post-process messages, defaults to ``temperature_softmax`` with a default temperature of 0.0
        :param post_func: a function to post-process messages, same as ``post_process_messages_func``
        :param config: a module to pass configurations, defaults to ``Config()``
        :param next_cache: customized method for next cache
        """
        self.has_init = True
        self.cache_enable_func = cache_enable_func
        self.pre_embedding_func = pre_func if pre_func else pre_embedding_func
        self.embedding_func = embedding_func
        self.data_manager: DataManager = data_manager
        self.similarity_evaluation = similarity_evaluation
        self.post_process_messages_func = post_func if post_func else post_process_messages_func
        self.config = config
        self.next_cache = next_cache

        @atexit.register
        def close():
            try:
                self.data_manager.close()
            except Exception as e:  # pylint: disable=W0703
                if not os.getenv("IS_CI"):
                    gptcache_log.error(e)

    def import_data(self, questions: List[Any], answers: List[Any], session_ids: Optional[List[Optional[str]]] = None) -> None:
        """Import data to GPTCache

        :param questions: preprocessed question Data
        :param answers: list of answers to questions
        :param session_ids: list of the session id.
        :return: None
        """

        self.data_manager.import_data(
            questions=questions,
            answers=answers,
            embedding_datas=[self.embedding_func(question) for question in questions],
            session_ids=session_ids if session_ids else [None for _ in range(len(questions))],
        )

    def flush(self):
        """Flush data, to prevent accidental loss of memory data,
        such as using map cache management or faiss, hnswlib vector storage will be useful
        """
        self.data_manager.flush()
        if self.next_cache:
            self.next_cache.data_manager.flush()

    @staticmethod
    def set_openai_key():
        import_openai()
        import openai  # pylint: disable=C0415

        openai.api_key = os.getenv("OPENAI_API_KEY")

    @staticmethod
    def set_azure_openai_key():
        import_openai()
        import openai  # pylint: disable=C0415

        openai.api_type = "azure"
        openai.api_key = os.getenv("OPENAI_API_KEY")
        openai.api_base = os.getenv("OPENAI_API_BASE")
        openai.api_version = os.getenv("OPENAI_API_VERSION")

cache = Cache()



from typing import List, Optional

import numpy as np

from gptcache.manager.vector_data.base import VectorBase, VectorData
from gptcache.utils import import_qdrant
from gptcache.utils.log import gptcache_log

import_qdrant()

# pylint: disable=C0413
from qdrant_client import QdrantClient
from qdrant_client.models import (
    PointStruct,
    HnswConfigDiff,
    VectorParams,
    OptimizersConfigDiff,
    Distance,
)


class QdrantVectorStore(VectorBase):
    """Qdrant Vector Store"""

    def __init__(
        self,
        url: Optional[str] = None,
        port: Optional[int] = 6333,
        grpc_port: int = 6334,
        prefer_grpc: bool = False,
        https: Optional[bool] = None,
        api_key: Optional[str] = None,
        prefix: Optional[str] = None,
        timeout: Optional[float] = None,
        host: Optional[str] = None,
        collection_name: Optional[str] = "gptcache",
        location: Optional[str] = "./qdrant",
        dimension: int = 0,
        top_k: int = 1,
        flush_interval_sec: int = 5,
        index_params: Optional[dict] = None,
    ):
        if dimension <= 0:
            raise ValueError(
                f"invalid `dim` param: {dimension} in the Qdrant vector store."
            )
        self._client: QdrantClient
        self._collection_name = collection_name
        self._in_memory = location == ":memory:"
        self.dimension = dimension
        self.top_k = top_k
        if self._in_memory or location is not None:
            self._create_local(location)
        else:
            self._create_remote(
                url, port, api_key, timeout, host, grpc_port, prefer_grpc, prefix, https
            )
        self._create_collection(collection_name, flush_interval_sec, index_params)

    def _create_local(self, location):
        self._client = QdrantClient(location=location)

    def _create_remote(
        self, url, port, api_key, timeout, host, grpc_port, prefer_grpc, prefix, https
    ):
        self._client = QdrantClient(
            url=url,
            port=port,
            api_key=api_key,
            timeout=timeout,
            host=host,
            grpc_port=grpc_port,
            prefer_grpc=prefer_grpc,
            prefix=prefix,
            https=https,
        )

    def _create_collection(
        self,
        collection_name: str,
        flush_interval_sec: int,
        index_params: Optional[dict] = None,
    ):
        hnsw_config = HnswConfigDiff(**(index_params or {}))
        vectors_config = VectorParams(
            size=self.dimension, distance=Distance.COSINE, hnsw_config=hnsw_config
        )
        optimizers_config = OptimizersConfigDiff(
            deleted_threshold=0.2,
            vacuum_min_vector_number=1000,
            flush_interval_sec=flush_interval_sec,
        )
        # check if the collection exists
        existing_collections = self._client.get_collections()
        for existing_collection in existing_collections.collections:
            if existing_collection.name == collection_name:
                gptcache_log.warning(
                    "The %s collection already exists, and it will be used directly.",
                    collection_name,
                )
                break
        else:
            self._client.create_collection(
                collection_name=collection_name,
                vectors_config=vectors_config,
                optimizers_config=optimizers_config,
            )

    def mul_add(self, datas: List[VectorData]):
        points = [
            PointStruct(id=d.id, vector=d.data.reshape(-1).tolist()) for d in datas
        ]
        self._client.upsert(
            collection_name=self._collection_name, points=points, wait=False
        )

    def search(self, data: np.ndarray, top_k: int = -1):
        if top_k == -1:
            top_k = self.top_k
        reshaped_data = data.reshape(-1).tolist()
        search_result = self._client.search(
            collection_name=self._collection_name,
            query_vector=reshaped_data,
            limit=top_k,
        )
        return list(map(lambda x: (x.score, x.id), search_result))

    def delete(self, ids: List[str]):
        self._client.delete(collection_name=self._collection_name, points_selector=ids)

    def rebuild(self, ids=None):  # pylint: disable=unused-argument
        optimizers_config = OptimizersConfigDiff(
            deleted_threshold=0.2, vacuum_min_vector_number=1000
        )
        self._client.update_collection(
            collection_name=self._collection_name, optimizer_config=optimizers_config
        )

    def flush(self):
        # no need to flush manually as qdrant flushes automatically based on the optimizers_config for remote Qdrant
        pass

    def close(self):
        self.flush()



from abc import ABCMeta, abstractmethod
from typing import Dict, Any


class ContextProcess(metaclass=ABCMeta):
    """ContextProcess: the context process interfacer, which is used to pre-process the lang conversation.
    By the way, the GPTCache will acquire more information and get a more accurate embedding vector.

    Example:
        .. code-block:: python

            from gptcache.processor.context import SummarizationContextProcess

            context_process = SummarizationContextProcess()
            cache.init(pre_embedding_func=context_process.pre_process)
    """

    @abstractmethod
    def format_all_content(self, data: Dict[str, Any], **params: Dict[str, Any]):
        """format all content of the llm request data as a string

        :param data: the user llm request data
        :type data: Dict[str, Any]
        """
        pass

    @abstractmethod
    def process_all_content(self) -> (Any, Any):
        """process all content of the llm request data, for extracting key information in context.
        In order to achieve this goal, you can pass the summary model and so on
        """
        pass

    def pre_process(self, data: Dict[str, Any], **params: Dict[str, Any]) -> (Any, Any):
        """ pre-process function, it's used as the GPTCache initialization param -- pre_embedding_func.

        :param data: the user llm request data
        :type data: Dict[str, Any]
        """
        self.format_all_content(data, **params)
        return self.process_all_content()



import numpy as np
import os

from gptcache.utils import import_fasttext
from gptcache.embedding.base import BaseEmbedding

import_fasttext()

import fasttext.util  # pylint: disable=C0413


class FastText(BaseEmbedding):
    """Generate sentence embedding for given text using pretrained models of different languages from fastText.

    :param model: model name, defaults to 'en'.
    :type model: str
    :param dim: reduced dimension of embedding. If this parameter is not provided, the embedding dimension (300) will not change.
    :type dim: int

    Example:
        .. code-block:: python

            from gptcache.embedding import FastText

            test_sentence = 'Hello, world.'
            encoder = FastText(model='en', dim=100)
            embed = encoder.to_embeddings(test_sentence)
    """

    def __init__(self, model: str = "en", dim: int = None):
        self.model_path = os.path.abspath(fasttext.util.download_model(model))
        self.ft = fasttext.load_model(self.model_path)

        if dim:
            fasttext.util.reduce_model(self.ft, dim)
        self.__dimension = self.ft.get_dimension()

    def to_embeddings(self, data, **_):
        """Generate embedding given text input

        :param data: text in string.
        :type data: str

        :return: a text embedding in shape of (dim,).
        """
        assert isinstance(data, str), "Only allow string as input."
        emb = self.ft.get_sentence_vector(data)
        return np.array(emb).astype("float32")

    @property
    def dimension(self):
        """Embedding dimension.

        :return: embedding dimension
        """
        return self.__dimension
